<?php // Please do not edit this file. If you do, save as UTF-8 without BOM!

/* Copyright 2010-2020 Martin Janecke <aneamal@prlbr.de>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


// declare (strict_types = 1); // only during development

namespace prlbr\aneamal;


// metadata types
const META_PLAIN = 1;   // plain variable without special function
const META_SPECIAL = 2; // variable with a special function in Aneamal
const META_CUSTOM = 4;  // variable that will be used as custom markup

// URI types
const URI_REMOTE = 1;   // absolute URI:  http://example.com/
const URI_DATA = 2;     // data URI:  data:text/plain;base64,Oik=
const URI_PATH = 4;     // local path:  /style/me.css  or  dog.nml
const URI_FILE = 8;     // relative to current file:  #section-2  or  ?v=27


/* Returns an HTML decimal numeric character reference for the given $byte, if
 * it is a printable ASCII character but not a blank space. Otherwise the $byte
 * is returned unchanged.
 */
function encode_printable_ascii (
	string $byte
):	string
{
	$ord = ord ($byte);
	return ($ord > 32 and $ord < 127)? '&#' . strval ($ord) . ';': $byte;
}


/* Encodes characters with a special meaning in HTML and also the dollar sign,
 * which is used as mark for math formulas by Aneamal, in a $string as HTML
 * character references. The $string is expected to be UTF-8 encoded; characters
 * that do not match UTF-8 are substituted by a Unicode replacement character.
 * The encoded $string is returned.
 */
function encode_special_chars (
	string $string
):	string
{
	return str_replace ('$', '&#36;', htmlspecialchars ($string, ENT_QUOTES | ENT_SUBSTITUTE, 'UTF-8'));
}


/* Runs a PHP file, providing it with parameters in the $data array. If the PHP
 * file returns an anonymous function, that function is cached, called and its
 * return value returned. If the PHP file returns something else, then that is
 * returned instead.
 */
function execute_module (
	string $filename,
	array $data
):	?string
{
	// cache for anonymous functions (= Closures) that are returned by modules
	static $closure = [];

	// execute cached modules immediately
	if (isset ($closure[$filename])):
		return (string) ($closure[$filename] ($data));
	endif;

	// otherwise return null, if the module file cannot be read; that state is
	// not cached by this script, because PHP itself caches is_readable!
	if (!is_readable ($filename)):
		return NULL;
	endif;

	// otherwise include the module file; parse errors are reported through an
	// anonymous function so that they will be cached and reported whenever an
	// erroneous module is used without PHP trying to parse it again and again
	try {
		$response = include_module ($filename, $data);
	} catch (\ParseError $e) {
		$response = function ($_) use ($e) {
			throw $e;
			return '';
		};
	}

	// if the module returned an anonymous function, cache and execute it
	if (is_object ($response) and is_a ($response, 'Closure')):
		$closure[$filename] = $response;
		return (string) ($response ($data));
	endif;

	// otherwise return the response of the extension file; this is mainly for
	// backwards compatibility as modules were not required to return a
	// cacheable anonymous function in the past
	return (string) $response;
}


/* Returns an array of two integers that correspond to two integer values given
 * as a $string and separated by a $delimiter. If the $delimiter is not found
 * in the $string, the second returned integer defaults to zero.
 */
function explode_integer_pair (
	string $delimiter,
	string $string
):	array
{
	$pair = explode ($delimiter, $string, 2);
	if (isset ($pair[1])):
		return [(int) $pair[0], (int) $pair[1]];
	else:
		return [(int) $pair[0], 0];
	endif;
}


/* Returns an array of substrings of $string formed by splitting $string at
 * $delimiter that is not protected with a backslash. The returned array will
 * contain a maximum of $limit elements with the last element containing the
 * rest of $string. Any $limit below 2 is treated as 1.
 */
function explode_unslashed (
	string $delimiter,
	string $string,
	int $limit = PHP_INT_MAX
):	array
{
	if ($limit < 2):
		return [$string];
	endif;
	$final = $limit - 1;
	$count = $offset = 0;
	$array = [];
	while ($count < $final and !is_null ($pos = strpos_unslashed ($string, $delimiter, $offset))):
		$array[$count++] = substr ($string, $offset, $pos - $offset);
		$offset = $pos + 1;
	endwhile;
	$array[] = substr ($string, $offset);
	return $array;
}


/* Returns a text that is composed from given $lines from the file referenced
 * by $filename. $lines is a string of comma-separated integers and ranges; a
 * range consists of two integers seperated by a colon. Here's an example:
 * "1,6,9:11,-3". This example would compose a text from the 1st, 6th, 9th,
 * 10th, 11th line and the 3rd last line from the file. The function returns the
 * whole text if the $lines parameter is not provided.
 */
function file_get_lines (
	string $filename,
	string $lines = NULL
):	?string
{
	if (!is_readable_file ($filename)):
		return NULL;
	endif;

	// return whole file, if no line choice was made
	if ($lines === NULL):
		$text = file_get_contents ($filename);
		return $text === false? NULL: $text;
	endif;

	// load lines of the file
	$text = file ($filename, FILE_IGNORE_NEW_LINES);
	if (!is_array ($text)):
		return NULL;
	endif;
	$count = count ($text);
	$output = [];

	foreach (explode (',', $lines) as $item):
		if ($item === ''):
			continue;
		endif;

		$range = explode (':', $item, 2);
		if (isset ($range[1])):
			$start = get_line_index ((int) $range[0], $count);
			$until = get_line_index ((int) $range[1], $count);
			if ($until >= $start):
				$start = max ($start, 0);
				$until = min ($until, $count - 1);
				for ($i = $start; $i <= $until; ++$i):
					$output[] = $text[$i];
				endfor;
			else:
				$start = min ($start, $count - 1);
				$until = max ($until, 0);
				for ($i = $start; $i >= $until; --$i):
					$output[] = $text[$i];
				endfor;
			endif;
		else:
			$i = get_line_index ((int) $item, $count);
			if (isset ($text[$i])):
				$output[] = $text[$i];
			endif;
		endif;
	endforeach;

	return implode ("\n", $output);
}


/* Expects a $directory, for example: '/foo/bar/baz'. Returns an array of the
 * given directory and all parent directories up to the root, for example
 * ['/foo/bar/baz', '/foo/bar', '/foo', '']. The function does only string
 * manipulation. It does not check the existence of directories and while it
 * resolves and hence ignores a path component '.', it does not resolve the
 * parent directory identifier '..'.
 */
function get_directories (
	string $directory
):	array
{
	$dir = '';
	$directories = [''];
	foreach (explode ('/', trim ($directory, '/')) as $segment):
		if ($segment !== '' and $segment !== '.'):
			$dir .= '/' . $segment;
			$directories[] = $dir;
		endif;
	endforeach;
	return array_reverse ($directories);
}


/* Returns the index from an array of a given $length (created with PHP's file
 * function) which corresponds to a given $line number (in the file). The line
 * number input can be negative; in that case lines are counted from the end of
 * the file. Note that the returned index will be outside the index range of the
 * array, iff the input line does not exist in the file.
 */
function get_line_index (
	int $line,
	int $length
):	int
{
	return $line + ($line < 0? $length: -1);
}


/* Classifies an Aneamal meta variable $key as one of three types and returns an
 * integer code for the corresponding type.
 */
function get_meta_type (
	string $key
):	int
{
	if (!isset ($key[0])):
		return 0;
	elseif ($key[0] === '&'):
		return META_CUSTOM;
	elseif (is_made_of ($key, 'abcdefghijklmnopqrstuvwxyz', '-0123456789abcdefghijklmnopqrstuvwxyz')):
		return META_SPECIAL;
	else:
		return META_PLAIN;
	endif;
}


/* Classifies an $uri as one of a few types of absolute and relative URIs and
 * returns an integer code for the corresponding type.
 */
function get_uri_type (
	string $uri
):	int
{
	if ($uri === ''):
		return 0;
	elseif (substr ($uri, 0, 2) === '//'):
		return URI_REMOTE;
	elseif ($uri[0] === '/'):
		return URI_PATH;
	elseif ($uri[0] === '?' or $uri[0] === '#'):
		return URI_FILE;
	elseif ($scheme = parse_url ($uri, PHP_URL_SCHEME)):
		return strtolower ($scheme) === 'data'? URI_DATA: URI_REMOTE;
	else:
		return URI_PATH;
	endif;
}


/* Includes a PHP file, providing parameters in the $_ array, and forwards the
 * included file's return value. NOTE: DO NOT CHANGE THE PARAMETER NAMES OF THIS
 * FUNCTION. Scripts included via this function could rely on them. The purpose
 * of this otherwise trivial function is to create a local variable scope for
 * the included file.
 */
function include_module (
	string $extension,
	array $_ = []
)	// should return a Closure or a string
{
	return include $extension;
}


/* Returns whether a given $string matches a simple pattern. The exact behaviour
 * depends on the number of provided parameters: If only $string and $charlist_1
 * are given, the function returns true iff $string contains only characters
 * that are also in $charlist_1. An empty $string results in true. If $string,
 * $charlist_1 and $charlist_2 are given, the function returns true iff the
 * first byte of $string is also in $charlist_1 and all further characters of
 * $string are also in $charlist_2. An empty $string results in false.
 */
function is_made_of (
	string $string,
	string $charlist_1,
	string $charlist_2 = NULL
):	bool
{
	if ($charlist_2 === NULL):
		return strspn ($string, $charlist_1) === strlen ($string);
	else:
		return
			isset ($string[0])
			and strpos ($charlist_1, $string[0]) !== false
			and strspn ($string, $charlist_2, 1) === strlen ($string) - 1
		;
	endif;
}


/* Returns whether a given $filename identifies an existing file, not a
 * directory, and is readable.
 */
function is_readable_file (
	string $filename
):	bool
{
	return is_file ($filename) and is_readable ($filename);
}


/* Returns whether a byte whose position in a $string is given by $pos (starting
 * at 0 for the first byte) is preceded by a backslash that is not preceded by
 * a backslash itself. The byte is slashed, iff the number of preceding
 * backslahes is odd.
 */
function is_slashed (
	string $string,
	int $pos
):	bool
{
	if (!isset ($string[$chk = $pos - 1])):
		return false;
	endif;
	while ($chk >= 0 and $string[$chk] === '\\') --$chk;
	return ($chk & 1) === ($pos & 1);
}


/* Parses a file token which may be found at the beginning of Aneamal blocks and
 * looks like [type:argument] where the :argument part is optional and the
 * case-insensitive type is converted to small letters. If the first byte of the
 * type is "a", "x" or "t" and the second byte is "-", the type is further
 * separated into supertype-subtype .
 * For historical reasons a colon immediately after the type identifiers "a",
 * "t", "x" is interpreted as a hyphen too. This implies that the these types
 * can only have arguments if they have a subtype in the first place.
 * The $token must be provided to the function WITHOUT surrounding square
 * brackets. Here are valid example tokens:
 *   Just a type:  d
 *   Type with subtype:  a-warning
 *   Type with argument:  i:Five foxes frolic freely.
 *   Type with subtype and argument:  t-excel:sheet 2
 * The function returns an array [(super)type, subtype, argument] where each
 * item is a string or NULL if missing.
 */
function parse_file_token (
	string $token = ''
):	array
{
	if ($token === ''):
		return [NULL, NULL, NULL];
	endif;

	$initial = strtolower ($token[0]);

	if (!isset ($token[1])):
		return [$initial, NULL, NULL];

	elseif ($token[1] === '-' and in_array ($initial, ['a', 't', 'x'], true)):
		if ($colon = strpos ($token, ':', 2)):
			return [$initial, mb_strtolower (substr ($token, 2, $colon - 2), 'UTF-8'), substr ($token, $colon + 1)];
		else:
			return [$initial, mb_strtolower (substr ($token, 2), 'UTF-8'), NULL];
		endif;

	elseif ($token[1] === ':'):
		if (in_array ($initial, ['a', 't', 'x'], true)):
			return parse_file_token ($initial . '-' . substr ($token, 2));
		else:
			return [$initial, NULL, substr ($token, 2)];
		endif;

	else:
		if ($colon = strpos ($token, ':', 2)):
			return [strtolower (substr ($token, 0, $colon)), NULL, substr ($token, $colon + 1)];
		else:
			return [strtolower ($token), NULL, NULL];
		endif;
	endif;
}


/* Prepares a barely formatted text $string for use as a HTML attribute value
 * and returns it. Barely formatted means that the only character with a special
 * meaning is the backslash. A backslash before a line break implies that the
 * line break shall be removed. A backslash before any other character implies
 * that the following character has no special meaning.
 */
function prepare_html_attribute (
	string $string
):	string
{
	return encode_special_chars (strip_slashes (trim (strip_slashed_breaks ($string))));
}


/* Returns an image resource which contains a resized version of an $original
 * image resource. The parameters $bv and $hv indicate the desired width and
 * height in pixels. A positive value will be fulfilled exactly, a negative
 * value is interpreted as maximum which is undershot iff the original image is
 * smaller or it is necessary to retain the image ratio. A value of zero is
 * interpreted as no restriction. Iff the image ratio cannot be retained, a
 * central crop of the original image is used. A more detailed explanation in
 * German can be found at: https://prlbr.de/2016/vorschau-bilder-berechnen/
 */
function resize_image (
	$original,
	int $bv,
	int $hv
)	// returns an image resource
{
	// Get the width (b) and height (h) from the original image.
	$bo = imagesx ($original);
	$ho = imagesy ($original);
	if ($bo < 1 or $ho < 1):
		return NULL;
	endif;

	// Set the width and height of the crop we are going to use from the
	// original image to the original width and height; i.e. we use the whole
	// image. We will change this in a few cases later.
	$ba = $bo;
	$ha = $ho;

	// Now calculate the width and height of the preview image, distinguishing
	// nine cases ...
	if ($bv < 0 and $hv < 0):
		$z = min (-$bv / $bo, -$hv / $ho, 1);
		$bn = roundp ($bo * $z);
		$hn = roundp ($ho * $z);
	elseif ($bv < 0 and $hv === 0):
		$bn = min ($bo, -$bv);
		$hn = roundp ($ho * $bn / $bo);
	elseif ($bv < 0 and $hv > 0):
		$hn = $hv;
		$bn = min (roundp ($bo * $hn / $ho), -$bv);
		$z = min ($bo / $bn, $ho / $hn);
		$ba = roundp ($bn * $z);
		$ha = roundp ($hn * $z);
	elseif ($bv === 0 and $hv < 0):
		$hn = min ($ho, -$hv);
		$bn = roundp ($bo * $hn / $ho);
	elseif ($bv === 0 and $hv === 0):
		$bn = $bo;
		$hn = $ho;
	elseif ($bv === 0 and $hv > 0):
		$hn = $hv;
		$bn = roundp ($bo * $hn / $ho);
	elseif ($bv > 0 and $hv < 0):
		$bn = $bv;
		$hn = min (roundp ($ho * $bn / $bo), -$hv);
		$z = min ($bo / $bn, $ho / $hn);
		$ba = roundp ($bn * $z);
		$ha = roundp ($hn * $z);
	elseif ($bv > 0 and $hv === 0):
		$bn = $bv;
		$hn = roundp ($ho * $bn / $bo);
	elseif ($bv > 0 and $hv > 0):
		$bn = $bv;
		$hn = $hv;
		$z = min ($bo / $bn, $ho / $hn);
		$ba = roundp ($bn * $z);
		$ha = roundp ($hn * $z);
	endif;

	// Create the preview image
	$resized = imagecreatetruecolor ($bn, $hn);
	if ($resized === false):
		return NULL;
	endif;

	// Color the preview image white so that we get a white background where the
	// original image may have been transparent.
	imagefill ($resized, 0, 0, imagecolorallocate ($resized, 255, 255, 255));

	// Calculate the starting point in the original image for the copying
	// process. This is usually (0, 0) unless we need to cut the image because
	// the image ratio of the preview image can't match the original image.
	$xa = (int) round (($bo - $ba) / 2);
	$ya = (int) round (($ho - $ha) / 2);

	// Create and return the preview image
	imagecopyresampled ($resized, $original, 0, 0, $xa, $ya, $bn, $hn, $ba, $ha);
	return $resized;
}


/* Returns the nearest positive integer for a given $real number.
 */
function roundp (
	float $real
):	int
{
	return max ((int) round ($real), 1);
}


/* Returns an array with three string components: all parts of an URI from
 * scheme to path, query, fragment.
 */
function split_uri_tail (
	string $uri
):	array
{
	// get fragment
	$delimiter = strpos ($uri, '#');
	if ($delimiter === false):
		$fragment = NULL;
	else:
		$fragment = substr ($uri, $delimiter + 1);
		$uri = substr ($uri, 0, $delimiter);
	endif;

	// get query
	$delimiter = strpos ($uri, '?');
	if ($delimiter === false):
		$query = NULL;
	else:
		$query = substr ($uri, $delimiter + 1);
		$uri = substr ($uri, 0, $delimiter);
	endif;

	return [$uri, $query, $fragment];
}


/* Strips the suffix .nml from an $url and returns the result. The suffix is the
 * last part of the $url before the first ? or # character.
 */
function strip_nml_suffix (
	string $url
):	string
{
	$length = strcspn ($url, '?#');
	$path = $length < strlen ($url)? substr ($url, 0, $length): $url;
	return substr ($path, -4) === '.nml'? substr ($path, 0, -4) . substr ($url, $length): $url;
}


/* Returns the portion of a given $path which is relative to the Aneamal root
 * directory or, if this is not possible, just the basename. This is used to
 * reduce the information published publicly in error messages.
 */
function strip_root (
	string $path
):	string
{
	$aneamal_root = dirname (__DIR__);
	if (strpos ($path, $aneamal_root) === 0):
		return substr ($path, strlen ($aneamal_root));
	else:
		return basename ($path);
	endif;
}


/* Removes all line feed characters U+000A that are slashed and the backslash
 * from a $string and returns the result.
 */
function strip_slashed_breaks (
	string $string
):	string
{
	if (strpos ($string, "\n") !== false):
		$lines = explode ("\n", $string);
		$final = array_pop ($lines);
		$string = '';
		foreach ($lines as $line):
			$string .= is_slashed ($line, strlen ($line))? substr ($line, 0, -1): $line . "\n";
		endforeach;
		$string .= $final;
	endif;
	return $string;
}


/* Removes all backslashes from a $string that are not slashed themselves and
 * returns the result.
 */
function strip_slashes (
	string $string
):	string
{
	$return = '';
	for ($i = 0, $length = strlen ($string); $i < $length; ++$i):
		if ($string[$i] !== '\\' or ++$i !== $length):
			$return .= $string[$i];
		endif;
	endfor;
	return $return;
}


/* Works like PHP's strpos function but ignores finds that are slashed with a
 * backslash or included within quotation marks or in a sub-context and returns
 * NULL instead of false when nothing is found.
 *
 * $haystack:  the string to search in
 * $needle:    the string to search for
 * $masks:     an array that defines mask marks where the keys provide the
 *             opening marks and the values provide the closing marks, e.g.
 *             ['"' => '"', "'" => "'"]; we do not look
 *             for the needle inside a quotation and we do not look for other
 *             quotation marks inside a quotation either
 * $offset:    initial search offset, 0 is the first byte in $haystack
 * $context:   a string such as ` which opens and closes a sub-string where we
 *             do not look for the needle, but where we do respect quotations
 *
 * returns:    the first position of an unslashed, unmasked needle in haystack
 *             at/after the initial offset or NULL if none was found
 */
function strpos_unmasked (
	string $haystack,
	string $needle,
	array $masks,
	int $offset = 0,
	string $context = ''
):	?int
{
	$haystack_length = strlen ($haystack);

	// build an array of opening quotation marks and add the context mark to it
	$opening_marks = array_keys ($masks);
	if ($context !== ''):
		$opening_marks[] = $context;
	endif;

	$needle_pos = -1;
	while (true):
		// find occurrence of an unprotected needle or return NULL
		if ($offset > $needle_pos and is_null ($needle_pos = strpos_unslashed ($haystack, $needle, $offset))):
			return NULL;
		endif;

		// search haystack, starting at offset, for the first occurrence of an
		// opening mark which is not protected by a backslash. The offset is
		// then set to that position and the found mark is saved. In case of no
		// find the new offset will be the haystack length.
		$mark_pos = $haystack_length;
		foreach ($opening_marks as $i => $candidate):
			if (is_null ($new_pos = strpos_unslashed ($haystack, $candidate, $offset))):
				unset ($opening_marks[$i]); // don't look for this again
			elseif ($new_pos < $mark_pos):
				$mark_pos = $new_pos;
				$mark = $candidate;
			endif;
		endforeach;

		// return needle pos if it's before/an unprotected opening mark
		if ($needle_pos <= $mark_pos):
			return $needle_pos;
		endif;

		// else find closing context/quotation mark or return NULL
		if ($mark === $context):
			if (is_null ($offset = strpos_unmasked ($haystack, $mark, $masks, $mark_pos + strlen ($mark)))):
				return NULL;
			endif;
			$offset += strlen ($mark);
		else:
			if (is_null ($offset = strpos_unslashed ($haystack, $masks[$mark], $mark_pos + strlen ($mark)))):
				return NULL;
			endif;
			$offset += strlen ($masks[$mark]);
		endif;
	endwhile;
}


/* Works like strpos_unmasked but additionally takes nesting levels into
 * account. That means it finds the first needle that is neither slashed nor
 * masked nor matched by an unslashed opening nesting mark before it.
 *
 * $haystack:  the string to search in
 * $needle:    the string to search for, also a closing nesting mark
 * $match:     an opening nesting mark to match a subsequent needle
 * $masks:     an array that defines masking marks where the keys provide the
 *             opening marks and the values provide the closing marks, e.g.
 *             ['"' => '"', "'" => "'"]
 * $offset:    initial search offset, 0 is the first byte in $haystack
 * $context:   a string such as ` which opens and closes a sub-string where we
 *             do not look for the needle, but where we do respect masks
 *
 * returns:    the first position of an unslashed, unmasked, unmatched needle in
 *             haystack at/after the initial offset or NULL if none was found
 */
function strpos_unmatched (
	string $haystack,
	string $needle,
	string $match,
	array $masks,
	int $offset = 0,
	string $context = ''
):	?int
{
	$unmatched = 0;
	$open = -1;
	$matchlen = strlen ($match);
	while (true):
		// find occurrence of an unmasked needle or return NULL
		if (is_null ($pos = strpos_unmasked ($haystack, $needle, $masks, $offset, $context))):
			return NULL;
		endif;
		++$unmatched;
		// find the number of unmasked opening nesting marks before the needle
		if (isset ($open)):
			$open = $offset - $matchlen;
			while (!is_null ($open = strpos_unmasked ($haystack, $match, $masks, $open += $matchlen, $context)) and $pos - $open >= $matchlen):
				--$unmatched;
			endwhile;
		endif;
		// return position of the needle iff the needle was unmatched
		if ($unmatched > 0):
			return $pos;
		else:
			$offset = $pos + strlen ($needle);
		endif;
	endwhile;
}


/* Works similar to PHP's strpos function but ignores finds that are protected
 * by an unprotected backslash and returns NULL instead of false when nothing
 * is found.
 *
 * $haystack:  the string to search in
 * $needle:    the string to search for
 * $offset:    initial search offset, 0 is the first byte in $haystack
 *
 * returns:    the first position of unslashed needle in haystack but at/after
 *             the initial offset or NULL if it cannot be found
 */
function strpos_unslashed (
	string $haystack,
	string $needle,
	int $offset = 0
):	?int
{
	$pos = $offset - 1;
	while ($pos = strpos ($haystack, $needle, $pos + 1) and is_slashed ($haystack, $pos));
	return $pos === false? NULL: $pos;
}


/* Works like strpos_unmasked except for finding the last instead of the first
 * occurrence. Look at strpos_unmasked for a detailed description.
 * This function must search forwards in the string until all occurrences have
 * been identified. Searching backwards from the end is not an option due to
 * the asymmetric nature of masks.
 */
function strrpos_unmasked (
	string $haystack,
	string $needle,
	array $masks,
	int $offset = 0,
	string $context = ''
):	?int
{
	$occurence = NULL;
	while (!is_null ($pos = strpos_unmasked ($haystack, $needle, $masks, $offset, $context = ''))):
		$occurence = $pos;
		$offset = $pos + 1;
	endwhile;
	return $occurence;
}


/* Changes a $string so that all input bytes from the ASCII range are preserved
 * and byte sequences outside the range are replaced by an Unicode replacement
 * character.
 */
function substitute_nonascii (
	string $string
):	string
{
	$ascii = true;
	$result = '';
	for ($i = 0, $length = strlen ($string); $i < $length; ++$i):
		if (ord ($string[$i]) < 128):
			$result .= $string[$i];
			$ascii = true;
		elseif ($ascii === true):
			$result .= "\xEF\xBF\xBD";
			$ascii = false;
		endif;
	endfor;
	return $result;
}


/* Encodes a $string so that it can be used as HTML id and URL fragment and CSS
 * selector. The result contains only ASCII alphanumeric characters and the
 * hyphen -.
 */
function x_encode (
	string $input
):	string
{
	// return the empty string unaltered
	if ($input === ''):
		return '';
	endif;

	// The following foreach-loop examines the input string from left to right,
	// one byte after the other. The current byte is stored in the variable
	// $byte. The variable $last holds the last thing added to the output
	// string.
	$last = $output = '';

	foreach (str_split ($input) as $i => $byte):
		// alphanumeric ASCII characters except x and X are not encoded
		if (is_made_of ($byte, 'abcdefghijklmnopqrstuvwyzABCDEFGHIJKLMNOPQRSTUVWYZ0123456789')):
			$output .= $last = $byte;
		// many remaining printable ASCII characters are treated as hyphen
		elseif (is_made_of ($byte, "\t\n .,;:!?\"'()[]{}<>-+=/%")):
			if ($last !== '-'):
				$output .= $last = '-';
			endif;
		// most inline markup is ignored unless slashed, in which case it is
		// treated as hyphen or xhex-encoded in case of a star
		elseif (is_made_of ($byte, '*~_|$`#^&@\\')):
			if (is_slashed ($input, $i)):
				if ($byte === '*'):
					$output .= $last = 'x2a';
				elseif ($last !== '-'):
					$output .= $last = '-';
				endif;
			endif;
		// x, X are encoded as xhex'd x
		elseif ($byte === 'x' or $byte === 'X'):
			$output .= $last = 'x78';
		// non-printable ASCII and non-ASCII bytes are xhex'd, meaning a
		// lowercase two-byte hexadecimal code for the byte is appended to a
		// lowercase letter x.
		else:
			$ord = ord ($byte);
			$output .= $last = ($ord < 16? 'x0': 'x') . strtolower (dechex ($ord));
		endif;
	endforeach;

	// trailing and leading - are removed and the output returned
	return trim ($output, '-');
}


// Version: 27.1