<?php // Please do not edit this file. If you do, save as UTF-8 without BOM!

/* Copyright 2010-2020 Martin Janecke <aneamal@prlbr.de>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


// declare (strict_types = 1); // only during development

namespace prlbr\aneamal;

// dependency
require __DIR__ . '/func.php';


class nml2html {

/* Constants
 */

// identifiers for different kinds of Aneamal files
public const main = 0;      // file requested by reader, loaded via main.php
public const linked = 1;    // file linked to with file tag like [a]->…
public const embedded = 2;  // code block with tag [a] above or quotation block
private const template = 3; // file /aneamal/xy(/index).nml referenced by [a-xy]
private const layout = 4;   // @aside/@footer/@header.nml or explicitly declared
private const meta = 5;     // automatically used @meta.nml or declared as @meta

// style for error messages and default URI for more error feedback
private const error_base_url = 'https://aneamal.org/error/';
private const error_style = 'background:#FA0;color:#000;padding:1ex 1em';

// the directory where to save preview images; their jpeg quality, see
// http://php.net/manual/en/function.imagejpeg.php; and their default size
// setting, see resize_image in func.php for details
public const pixdir = '/aneamal/pix';
public const pixiq = 83;
public const pixsize = '-640,-640';

// letters and digits
private const alphanumeric = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789';

// maximum number of file_aneamal calls to prevent infinite recursion
private const max_inclusions = 256;

// default maximum size of text file included with [a], [h], [t], [b] etc.
private const default_textcap = 262144; // 2^18 Byte = 256 KiB

// string markup whose content is not interpreted as containing Aneamal markup;
// the array uses the format 'left mark' => 'right mark'
private const masks = ['|' => '|', '$' => '$', '{' => '}', '->`' => '`', '->!`' => '`'];

// brackets create strings that can be nested
private const brackets = ['(' => ')', '[' => ']'];

// heading/section-break marks and their corresponding levels
private const sectioners = ['===' => 1, '---' => 2, '- -' => 3];

// expandable-section heading marks and their corresponding levels
private const expanders = ['+++' => 2, '+ +' => 3];

// alignment mark => [kind, HTML classes]; different kinds can be combined
private const aligners = [
	':..' => ['align', '_align _left'],
	'..:' => ['align', '_align _right'],
	':.:' => ['align', '_align _justify'],
	'.:.' => ['align', '_align _center'],
	'::.' => ['align', '_align _mid _left'],
	'.::' => ['align', '_align _mid _right'],
	': :' => ['allot', '_columns'],
	':::' => ['allot', '_row'],
	':.'  => ['float', '_float _port'],
	'.:'  => ['float', '_float _starboard'],
	'::'  => ['clear', '_clear'],
];

// the character & followed by a letter or digit is a custom mark, i.e. it can
// be defined by the author; & followed by some non-alphanumeric characters is
// predefined as follows
private const default_markvars = [
	'+' => "<span style='font-size:larger'>",
	'-' => "<span style='font-size:smaller'>",
	',' => "<span style='font-weight:normal'>",
	';' => '</span>',
];

// metadata names for use with links as values which have no special function
// in Aneamal but are recognized by Aneamal and published as HTML <link>
private const recognized_html_links = [
	// a selection from W3C and WHATWG standards
  	'canonical' => "<link rel='canonical' href='%s'>",
  	'icon'      => "<link rel='icon' href='%s'>",
	'license'   => "<link rel='license' href='%s'>",
	'next'      => "<link rel='next' href='%s'>",
	'prev'      => "<link rel='prev' href='%s'>",
	// further meta links
  	'atom'		=> "<link rel='alternate' type='application/atom+xml' href='%s'>",
  	'me'        => "<link rel='me' href='%s'>",
  	'rss'		=> "<link rel='alternate' type='application/rss+xml' href='%s'>",
  	'shortlink' => "<link rel='shortlink' href='%s'>",
  	'up'        => "<link rel='up' href='%s'>",
];

// metadata names for use with text values which have no special function in
// Aneamal but are recognized by Aneamal and published as HTML <meta>
private const recognized_html_metas = [
	// a selection from W3C and WHATWG standards
	'author'      => "<meta name='author' content='%s'>",
	'description' => "<meta name='description' content='%s'>",
	'keywords'    => "<meta name='keywords' content='%s'>",
	// further meta names
	'publisher'   => "<meta name='publisher' content='%s'>",
	'robots'      => "<meta name='robots' content='%s'>",
	'viewport'    => "<meta name='viewport' content='%s'>",
];

// these metadata names may be declared more than once in the same file, while
// all others can only be declared once
private const metadata_multiples = [
	'htmlhead',
	'javascript',
	'script',
	'style',
	'stylesheet',
];

// these metadata names are aliases
private const metadata_aliases = [
	'classes'	=> 'class',
	'language'	=> 'lang',
];


/* Properties
 */

// holds error messages, gets cleared after metadata and each block is processed
private $errors = [];
// base URL for more information on error messages
private $errormore = NULL;
// filename + additional note for embedded content to tell where errors occurred
private $docname = '';

// local file path of the Aneamal root directory; no trailing slash
private $root = __DIR__ . '/..';
// Aneamal root directory relative to the domain in an URL; no trailing slash
private $url = '';
// Aneamal document's directory relative to the Aneamal root; no trailing slash
private $dir = '';

// working copy of the Aneamal document's lines
private $lines = [];

// metadata where the value is an URI; format 'name' => 'value'
private $linkvars = [];
// metadata where the value is text; format 'name' => 'value'
private $textvars = [];
// custom marks; format 'mark' => 'HTML content'
private $markvars = [];

// locally declared metadata names including custom marks
private $metadecs = [];

// URIs identifying translations of the document; format 'langcode' => 'URI'
private $altlangs = [];
// URIs to JavaScript files that shall be referenced
private $javascripts = [];
// URIs to CSS files that shall be referenced
private $stylesheets = [];

// URIs that are published in HTML <link rel> elements; format 'name' => 'URI'
private $links = [];
// texts that are published im HTML <meta name> elements; format 'name' => 'URI'
private $metas = [];

// HTML snippets to be put in the HTML <head>
private $metahtmls = [];
// JavaScript snippets to be put at the end of the HTML <body>
private $metascripts = [];
// CSS snippets to be put in the HTML <head>
private $metastyles = [];

// manually declared metadata filename; empty string means don't autoload either
private $meta = NULL;
// manually declared look filename; empty string means don't autoload either
private $look = NULL;
// manually declared header filename; empty string means don't autoload either
private $header = NULL;
// manually declared aside filename; empty string means don't autoload either
private $aside = NULL;
// manually declared footer filename; empty string means don't autoload either
private $footer = NULL;
// layout value 'manual' prevents autoloading header/aside/footer/look, 'blank'
// prevents any display, NULL is the default behaviour declared with 'auto'
private $layout = NULL;

// original character encoding declared s metadata
private $encoding = NULL;

// BCP47 language code for the document, but lowercased
private $language = NULL;
// value 'rtl' iff the document is written in right-to-left script
private $direction = NULL;

// value for the HTML class attribute
private $classes = '';
// value for the HTML role attribute; only used for linked/embedded documents
private $role = '';
// value for HTML <title>; only used for the main document
private $title = '';
// added at the end of the HTML <title>; possible use is branding via @meta.nml
private $titletail = '';

// image dimensions of generated preview for [j]-> inclusion in pixels: width
// and height separated by comma; positive values force the given size, negative
// values set a maximum; zero means that the value is not set
private $pixels = NULL;

// maximum size of text file included with [a], [h], [t], [b] etc. in bytes
private $textcap = NULL;

// tracks whether expandable sections of a given rank have been started
private $expandable = [2 => false, 3 => false];
// for HTML <details> and </details> tags that enclose an expandable section and
// are triggered by Aneamal headings/breaks, but do not belong to heading/break
private $sectiontags = '';


/* Public methods
 */

/* This is the only constructor of the class. It preprocesses Aneamal file which
 * includes processing sandwich markup, parsing/removing metadata, removing
 * comments and converting the document to UTF-8.
 *
 * $source:     the source text of an Aneamal file
 * $dir:        directory in which the parsed file is located relative to the
 *              Aneamal root, starting with a slash; needed to locate other
 *              files referenced from this
 * $url:        Aneamal root directory relative to the domain in an URL; an
 *              empty string means the Aneamal root is the document root
 * $docname:    identifier for this document in error messages
 * $kind:       an integer corresponding to the kind of Aneamal document, either
 *              self::main, self::embedded, self::linked, self::template,
 *              self::layout or self::meta
 * $linkvars:   metadata where the value is an URI; format 'name' => 'value'
 * $textvars:   metadata where the value is text; format 'name' => 'value'
 * $markvars:   custom marks; format 'mark' => 'HTML content'
 */
public function __construct (
	string $source,
	string $dir,
	string $url,
	string $docname = '',
	int $kind = self::main,
	array $linkvars = [],
	array $textvars = [],
	array $markvars = []
)	// returns nothing
{
	// remove optional UTF-8 byte order mark
	if (substr ($source, 0, 3) === "\xEF\xBB\xBF"):
		$source = substr ($source, 3);
	endif;

	// remove trailing slashes from the directories
	$dir = rtrim ($dir, '/');
	$url = rtrim ($url, '/');

	// check whether the alleged directory of the file exists
	if (!is_dir ($this->root . $dir)):
		trigger_error ('Invalid directory: ' . $dir, E_USER_ERROR);
		exit;
	endif;

	// $load_meta_url may later be changed by a @meta declaration in self::main
	// Aneamal files. An empty string would indicate that no @meta.nml file
	// shall be used automatically and a different string would point to a
	// manually chosen metadata file. But at the moment we assume no such
	// setting and will automatically load a @meta.nml file.
	$load_meta_url = NULL;

	// The following loop usually runs only once because we will break out of it
	// unless the file is not UTF-8-encoded or a @meta declaration is found that
	// contradicts the automatic handling of @meta.nml files in the first run.
	// Files that are not UTF-8 encoded will be converted to UTF-8, wrong
	// automatic meta handling will be reversed and the loop will run a second
	// time, if needed.
	foreach ([true, false] as $first):

		// local file path of the Aneamal root directory without /.. component
		$this->root = dirname (__DIR__);

		// save location data
		$this->dir = $dir;
		$this->url = $url;
		$this->docname = $docname;

		// no metadata has been loaded from @meta.nml files automatically in
		// this run of the loop yet, but we will keep track of it
		$loaded_meta_automatically = false;

		// load metadata from metadata files in case of self::main Aneamal
		// files; inherit metadata variables in case of other Aneamal files
		if ($kind === self::main):
			if ($load_meta_url === NULL):
				$loaded_meta_automatically = $this->load_meta ();
			elseif ($load_meta_url !== ''):
				$this->load_meta ($load_meta_url);
			endif;
		else:
			$this->linkvars = $linkvars;
			$this->textvars = $textvars;
			$this->markvars = $markvars;
		endif;

		// prepare $this->lines as array of all lines in the document, process
		// sandwich markup, process and remove metadata and comments
		$this->lines = explode ("\n", str_replace (["\r\n", "\r"], "\n", $source));
		$this->preprocess_lines ();
		$this->preprocess_comments_meta ();

		// now, after this Aneamal file's metadata has been processed, get
		// a possible @meta declaration
		if ($kind === self::main):
			$load_meta_url = $this->meta;
		endif;

		// the character encoding of embedded files has already been handled
		// when loading the containing file and, since embedded files do not
		// load metadata files, we break out of the loop for their kind
		if ($kind === self::embedded):
			is_null ($this->encoding) or $this->error ('@charset declared in embedded file or quotation block', 46);
			break;
		endif;

		// reduce the text to ASCII characters plus the unicode replacement
		// character, if the text did not match its alleged encoding or the
		// encoding is unknown to PHP
		$encoding = is_null ($this->encoding)? 'UTF-8': @mb_preferred_mime_name ($this->encoding);
		if ($first and !$encoding): // unknown encoding
			$this->reset ();
			$this->error ('Unrecognized character encoding', 3);
			$source = substitute_nonascii ($source);
		elseif ($first and !mb_check_encoding ($source, $encoding)): // alledged encoding incorrect
			$this->reset ();
			$this->error ('Character encoding mismatch: expected ' . $encoding, 2);
			$source = substitute_nonascii ($source);
		elseif ($first and $encoding !== 'UTF-8'): // encoding correct, but not UTF-8
			$this->reset ();
			mb_convert_variables ('UTF-8', $encoding, $source);
		elseif ($first and isset ($load_meta_url) and $loaded_meta_automatically || $load_meta_url !== ''):
			$this->reset ();
		else:
			// use default language settings if not set in meta; this does not
			// apply to embedded files, which inherit the settings of the
			// containing document in process_file_aneamal
			is_null ($this->language) and $this->language = '';
			is_null ($this->direction) and $this->direction = 'ltr';
			// no conversions done, so break out of the loop
			break;
		endif;

	endforeach;
}


/* Generates and returns the HTML <body> of the Aneamal file managed by this
 * object. It does so by gathering lines of the file that belong to one block
 * to be parsed by the block method. This method is also responsible for
 * printing error messages in between blocks.
 */
public function body (
):	string
{
	$body = $this->get_errors ('Errors in meta');
	$lines = $numbers = [];
	$this->sectiontags = '';

	foreach ($this->lines as $number => $line):
		if ($line !== ''):
			$lines[] = $line;
			$numbers[] = $number;
		elseif (!empty ($lines)):
			$block = $this->block ($lines); // sets $this->sectiontags
			$body .= $this->sectiontags . $block;
			$body .= $this->get_errors ('Errors in the previous block', $numbers);
			$lines = $numbers = [];
			$this->sectiontags = '';
		endif;
	endforeach;

	if (!empty ($lines)):
		$block = $this->block ($lines); // sets $this->sectiontags
		$body .= $this->sectiontags . $block;
		$body .= $this->get_errors ('Errors in the last block', $numbers);
	endif;

	return $body . $this->end_expandables ();
}


/* Generates and returns the whole HTML webpage representing the Aneamal file
 * managed by this object, including implied header/footer/etc. files, styles
 * and scripts.
 */
public function document (
):	string
{
	$header = $footer = $aside = '';

	// hidden document
	if ($this->layout === 'blank'):
		return '';
	endif;

	// automatic layout
	if ($this->layout !== 'manual'):

		// prepare a list of directories to look for layout elements in; it is
		// the current directory and all parent directories to the Aneamal root
		$directories = get_directories ($this->dir);

		// include header
		if ($this->header === NULL):
			foreach ($directories as $dir):
				$filename = $this->root . $dir . '/@header.nml';
				if (is_readable_file ($filename)):
					$header = $this->process_file_aneamal (file_get_contents ($filename), $dir, '@header.nml', self::layout, 'header');
					break;
				endif;
			endforeach;
		endif;

		// include aside
		if ($this->aside === NULL):
			foreach ($directories as $dir):
				$filename = $this->root . $dir . '/@aside.nml';
				if (is_readable_file ($filename)):
					$aside = $this->process_file_aneamal (file_get_contents ($filename), $dir, '@aside.nml', self::layout, 'aside');
					break;
				endif;
			endforeach;
		endif;

		// include footer
		if ($this->footer === NULL):
			foreach ($directories as $dir):
				$filename = $this->root . $dir . '/@footer.nml';
				if (is_readable_file ($filename)):
					$footer = $this->process_file_aneamal (file_get_contents ($filename), $dir, '@footer.nml', self::layout, 'footer');
					break;
				endif;
			endforeach;
		endif;

		// include look (implied stylesheet)
		if ($this->look === NULL):
			foreach ($directories as $dir):
				$canonical = $dir . '/@look.css';
				if (is_file ($this->root . $canonical)):
					array_unshift ($this->stylesheets, $this->url . $canonical);
					break;
				endif;
			endforeach;
		endif;

	endif;

	// manual layout: header
	if ($this->header !== NULL and $this->header !== ''):
		$header = $this->file_aneamal ($this->header, 'header');
	endif;

	// manual layout: aside
	if ($this->aside !== NULL and $this->aside !== ''):
		$aside = $this->file_aneamal ($this->aside, 'aside');
	endif;

	// manual layout: footer
	if ($this->footer !== NULL and $this->footer !== ''):
		$footer = $this->file_aneamal ($this->footer, 'footer');
	endif;

	// manual layout: look
	if ($this->look !== NULL and $this->look !== ''):
		array_unshift ($this->stylesheets, $this->look);
	endif;

	// get body; must be generated before the head, because the document title
	// published in the head may be generated from the main heading in the body
	$body = $header . '<main>' . $this->body () . '</main>' . $aside . $footer;

	// add javascript
	foreach (array_unique ($this->javascripts) as $val):
		$body .= "<script src='" . encode_special_chars ($val) . "'></script>";
	endforeach;
	foreach (array_unique ($this->metascripts) as $val):
		$body .= "<script>$val</script>";
	endforeach;

	// build <head> content
	$head = $this->head ();

	// <html> and <body> attributes
	$dir = $this->direction !== 'ltr'? " dir='{$this->direction}'": '';
	$lang = $this->language? " lang='{$this->language}'": '';
	$class = $this->classes === ''? '': " class='" . encode_special_chars ($this->classes) . "'";

	// build document
	return "<!doctype html><html{$dir}{$lang}><head>$head</head><body{$class}>$body</body></html>";
}


/* Private methods
 */

/* Categorizes an alignment $mark at the start of a block, which consists of
 * $lines, as one of four kinds align/allot/float/clear, translates it into a
 * class name and adds it as [kind, class] to the &$alignment array, which may
 * already be populated by such items. All the classes are applied to the
 * processed block, unless it was already done during the processing of the
 * block, and the result is returned.
 * $inside_note is only passed through this method without relevance here.
 * The order of arguments of this method is in accordance with function block.
 */
private function align (
	string $mark,
	array $lines,
	bool $inside_note,
	array &$alignment
)
{
	$length = strlen ($mark);

	if (isset ($lines[0][$length]) and !is_made_of ($lines[0][$length], "\t\x20")):
		$this->error ('Missing whitespace after alignment mark ' . $mark, 211);
	endif;

	// remove alignment mark
	$lines[0] = ltrim (substr ($lines[0], $length));
	$lines[0] === '' and array_shift ($lines);

	// the clear mark is the only alignment mark that can stand on its own
	if (empty ($lines)):
		if ($mark === '::'):
			return "<div class='" . self::aligners['::'][1] . "'></div>";
		else:
			$this->error ('Aligned block missing: expected content below ' . $mark, 208);
			return '';
		endif;
	endif;

	// prevent incompatibles like left- and right-aligned; prepare alignment
	if (isset ($alignment[self::aligners[$mark][0]])):
		$this->error ('Alignment of this kind already set: ' . self::aligners[$mark][0], 213);
	else:
		$alignment[self::aligners[$mark][0]] = self::aligners[$mark][1];
	endif;

	// process the aligned block; it may apply the alignment itself, e.g. if it
	// starts with another alignment mark or is an expandable-section heading
	$content = $this->block ($lines, $inside_note, $alignment);

	// apply alignment, if not done by the block itself
	if (!empty ($alignment)):
		$content = "<div class='" . implode (' ', $alignment) . "'>$content</div>";
		$alignment = [];
	endif;

	return $content;
}


/* Processes $lines of an Aneamal document which together form an Aneamal block
 * and returns the corresponding HTML code. The processing depends on the type
 * of block. The type is identified by block markup, usually to be found at the
 * start of the block. Headings are only allowed, if not $inside_note.
 * &$alignment consists of arrays [kind => class] and is only passed through
 * this method without relevance here; see function align.
 */
private function block (
	array $lines,
	bool $inside_note = false,
	array &$alignment = []
):	string
{
	// join the lines of the block and extract the first two and three chars
	$block = implode ("\n", $lines);
	$three = substr ($block, 0, 3);
	$two   = substr ($three, 0, 2);

	// standard headings and section breaks
	if (array_key_exists ($three, self::sectioners)):
		if ($inside_note):
			$this->error ('Heading or section break in notes', 212); // also used below
		else:
			return $this->sectioner ($block);
		endif;
	endif;

	// expandable-section headings and breaks
	if (array_key_exists ($three, self::expanders)):
		if ($inside_note):
			$this->error ('Heading or section break in notes', 212); // also used above
		else:
			return $this->expander ($block, $alignment);
		endif;
	endif;

	// notes consist of three or more underscores and a block below
	if ($three === '___'):
		if (!is_made_of ($lines[0], '_')):
			$this->error ('Invalid notes markup: expected only underscores in the first line', 75);
			return '';
		elseif (!isset ($lines[1])):
			$this->error ('Note missing: expected content below ___', 76);
			return '';
		else:
			array_shift ($lines);
			return "<div role='note'><hr>" . $this->block ($lines, true) . '</div>';
		endif;
	endif;

	// math block: starts and ends with double dollar signs
	if ($two === '$$'):
		return $this->math_block ($block);
	endif;

	// code block: every line starts with a vertical bar
	if ($lines[0][0] === '|'):
		foreach ($lines as $n => &$line):
			if ($line[0] === '|'):
				$line = substr ($line, 1);
			else:
				return $this->embedded_file (implode ("\n", array_slice ($lines, 0, $n)), '', implode ("\n", array_slice ($lines, $n)));
			endif;
		endforeach;
		return $this->embedded_file (implode ("\n", $lines));
	endif;

	// bulleted list, optionally indented with dots and whitespace
	if ($two === '<>' && $three !== '<><' or $lines[0][0] === '.' && substr (ltrim ($lines[0], ".\t\x20"), 0, 2) === '<>'):
		return $this->bulleted_list ($lines);
	endif;

	// tagged list
	if ($block[0] === '<'):
		return $this->tagged_list ($block);
	endif;

	// file candidate, i.e. file or paragraph with bracketed string at its start
	if ($lines[0][0] === '[' and $offset = strpos_unslashed ($block, ']')):
		return $this->file ($block, $offset);
	endif;

	// block quotation
	if ($lines[0][0] === '>'):
		$text = $citation = '';
		foreach ($lines as $n => $line):
			if ($line[0] === '>'):
				$text .= substr ($line, 1) . "\n";
			else:
				$citation = implode ("\n", array_slice ($lines, $n));
				break;
			endif;
		endforeach;
		return $this->process_file_aneamal (
			substr ($text, 0, -1),
			$this->dir,
			$this->docname . ' > quotation block',
			self::embedded,
			'blockquote',
			NULL,
			$citation
		);
	endif;

	// alignment
	if (in_array ($three, [': .', '. :'], true)):
		$this->error ("Reserved characters $three at paragraph start", 209);
	elseif (array_key_exists ($three, self::aligners)):
		return $this->align ($three, $lines, $inside_note, $alignment);
	elseif (array_key_exists ($two, self::aligners)):
		return $this->align ($two, $lines, $inside_note, $alignment);
	endif;

	// numbered list
	if ($pos = strcspn ($lines[0], "\x20\t") and $lines[0][$pos - 1] === '.'):
		if (!empty ($this->item_number (substr ($lines[0], 0, $pos), []))):
			return $this->numbered_list ($lines);
		endif;
	endif;

	// multiple-choice question
	foreach ($lines as $n => $line):
		if ($line[0] === '{'):
			if ($n === 0):
				return $this->multiple_choice ($block);
			else:
				return $this->multiple_choice (
					implode ("\n", array_slice ($lines, $n)),
					implode ("\n", array_slice ($lines, 0, $n))
				);
			endif;
		endif;
	endforeach;

	// simple paragraph
	return '<p>' . $this->phrase ($block) . '</p>';
}


/* Finds the right bracket in a $string that corresponds to the left bracket
 * whose position in the $string is initially given by &$index. If a
 * corresponding right bracket is found, &$index is set to its position and the
 * part of the $string between the brackets is returned. If no corresponding
 * bracket is found, NULL is returned as signal that the left bracket does not
 * mark a bracketed string.
 */
private function bracketed (
	string $string,
	int &$index
):	?string
{
	// get the left bracket
	$left = $string[$index];

	// find the corresponding right bracket
	$pos = strpos_unmatched ($string, self::brackets[$left], $left, self::masks, $index + 1, '`');

	if ($pos === NULL):
		return NULL;
	endif;

	$group = substr ($string, $index + 1, $pos - $index - 1);
	$index = $pos;

	// remove slashed linebreaks and return the result
	return strip_slashed_breaks ($group);
}


/* Parses $lines that together form a bulleted list and returns it as HTML <ul>
 * list. Bulleted list items may be indented. The indentation level is marked by
 * dots. Space/tabs are allowed between the dots without having meaning.
 */
private function bulleted_list (
	array $lines
):	string
{
	$list = $item = '';
	$prelevel = $level = -1;

	foreach ($lines as $line):
		
		$bullet = strpos ($line, '<>');
		if ($bullet === false):
			$item .= "\n" . $line;
			continue;
		endif;

		$indent = substr ($line, 0, $bullet);
		if (strspn ($indent, ".\t\x20") < $bullet):
			$item .= "\n" . $line;
			continue;
		endif;

		$list .= $this->item ($item);
		$level = substr_count ($indent, '.');

		if ($level === $prelevel):
			$list .= '</li><li>';
		elseif ($level < $prelevel):
			$list .= str_repeat ('</li></ul>', $prelevel - $level) . '<li>';
		else:
			$list .= str_repeat ("<ul><li class='_skip'>", $level - $prelevel - 1);
			$list .= '<ul><li>';
		endif;

		$item = ltrim (substr ($line, $bullet + 2));
		$prelevel = $level;
	endforeach;

	return $list . $this->item ($item) . str_repeat ('</li></ul>', $level + 1);
}


/* Prepares $code for output as a HTML <code> element.
 */
private function code (
	string $code
):	string
{
	return '<code>' . str_replace ("\n", '<br>', encode_special_chars (strip_slashes ($code))) . '</code>';
}

/* Finds the right cross -+ in a $string that corresponds to the left cross +-
 * whose position in the $string is initially given by &$index. Then &$index is
 * set to the position of the + in the right cross (or to the last character,
 * if no right cross is found) and the part of the $string between the crosses
 * is returned.
 */
private function crossedout (
	string $string,
	int &$index
):	?string
{
	// get the left bracket
	$left = $string[$index];

	// find the corresponding right cross
	$pos = strpos_unmatched ($string, '-+', '+-', self::masks, $index + 2, '`');

	if ($pos === NULL):
		$group = substr ($string, $index + 2);
		$index = strlen ($string) - 1;
		$this->error ('Crossed out string not closed: expected -+', 206);
	else:
		$group = substr ($string, $index + 2, $pos - $index - 2);
		$index = $pos + 1;
	endif;

	// remove slashed linebreaks and return the result
	return strip_slashed_breaks ($group);
}


/* Decodes data from a base64-encoded data $uri and returns it or NULL on error.
 * Although this method does not validate the data URI againt the data URI
 * specification, it can be used to check if the data URI fulfills criteria
 * relevant to Aneamal.
 */
private function data_uri (
	string $uri,
	bool $convert_to_utf8 = false
):	?string
{
	// default data URI text encoding
	$encoding = 'US-ASCII';

	// find the colon (end of URI scheme) and first comma (start of data)
	$colon = strpos ($uri, ':');
	$comma = stripos ($uri, ';base64,', $colon);
	if (!$colon or !$comma):
		$this->error ('Invalid or not base64-encoded data URI', 61);
		return NULL;
	endif;

	// split the meta information into its components
	$about = explode (';', substr ($uri, $colon + 1, $comma - $colon - 1));

	// fetch the data from the data URI
	$data = rawurldecode (substr ($uri, $comma + 8));

	// base64-decode the data
	$data = base64_decode (chunk_split (str_replace ("\n", '', $data)), true);
	if ($data === false):
		$this->error ('Corrupt base64 encoding', 62);
		return NULL;
	endif;

	// check if data is present
	if ($data === ''):
		$this->error ('Data missing in data URI', 63);
		return NULL;
	endif;

	// convert the data to the output character encoding UTF-8
	if ($convert_to_utf8):
		// find the data URI's character encoding
		foreach ($about as $parameter):
			$component = explode ('=', rawurldecode ($parameter), 2);
			if (strtolower ($component[0]) === 'charset' and isset ($component[1])):
				$encoding = str_replace ('"', '', $component[1]);
				break;
			endif;
		endforeach;

		// check whether the encoding is known to PHP and transform the name
		// of the encoding to its preferred mime name
		if ($encoding = @mb_preferred_mime_name ($encoding)):
			// check whether the text matches its alleged encoding
			if (mb_check_encoding ($data, $encoding)):
				// convert the text to the output encoding UTF-8
				if ($encoding !== 'UTF-8'):
					mb_convert_variables ('UTF-8', $encoding, $data);
				endif;
			else:
				$this->error ('Character encoding mismatch: expected ' . $encoding, 64);
				$data = substitute_nonascii ($data);
			endif;
		else:
			$this->error ('Unrecognized character encoding', 65);
			$data = substitute_nonascii ($data);
		endif;
	endif;

	return $data;
}


/* A target at the start of a quotation block's or file's caption or math block
 * label is used as identifier for the whole quotation block/file/math block.
 * This method recognizes a target by the character # at the start of a given
 * &$caption. If it exists, the target is detached from the &$caption, processed
 * and returned as HTML id attribute with leading space. If no target exists, an
 * empty string is returned.
 */
private function detach_caption_target (
	string &$caption
):	string
{
	if ($caption[0] !== '#'):
		return '';
	endif;

	if (!isset ($caption[1])):
		$this->error ('Target text missing after initial #', 177);
		return $caption = '';
	endif;

	$index = 1;
 	if ($caption[1] === '{'):
		$name = $this->fragment_identifier ($this->encurled ($caption, $index), true);
		$caption = ltrim (substr ($caption, $index + 1));
	else:
		$name = $this->fragment_identifier ($this->targeted ($caption, $index), true);
		$caption = ltrim (substr ($caption, 1));
	endif;

	return $name === ''? '': " id='$name'";
}


/* Prepares the $content of an embedded file with the right method for its type,
 * which is encoded in the file $token. See function parse_file_token in
 * func.php for a description of the $token. If the optional $caption is
 * provided, the processed content and caption are returned as an HTML <figure>,
 * otherwise the content is returned without <figure> element.
 */
private function embedded_file (
	string $content,
	string $token = '',
	string $caption = ''
):	string
{
	list ($type, $subtype, $argument) = parse_file_token ($token);

	if (isset ($argument) and in_array ($type, ['a', 'b', 'd', 'h', 'p', 'q'], true)):
		$this->error ('Unsupported argument for embedded file of type: ' . $type, 159);
		return '';
	endif;

	// call the right method to process the given file type
	if ($type === NULL):
		$file = '<pre><code>' . encode_special_chars ($content) . '</code></pre>';
	elseif ($type === 'a'):
		$file = $this->process_file_aneamal ($content, $this->dir, $this->docname . ' > embedded file', self::embedded, 'div', $subtype);
	elseif ($type === 'b'):
		$file = $this->process_file_tsv ($content);
	elseif ($type === 'd'):
		$file = $this->process_file_tsv ($content, true);
	elseif ($type === 'h'):
		$file = $content;
	elseif ($type === 'p'):
		$file = $this->process_file_tsv ($content, false, true);
	elseif ($type === 'q'):
		$file = $this->process_file_tsv ($content, true, true);
	elseif ($type === 't'):
		$file = $this->process_file_text ($content, $subtype, $argument);
	elseif (in_array ($type, ['i', 'j', 'v', 'w', 'x'], true)):
		$this->error ('Unsupported file type: ' . $type, 145);
		return '';
	else:
		$this->error ('Unrecognized file type: ' . $type, 146);
		return '';
	endif;

	// handle the optional caption and id
	$id = '';
	if ($caption !== ''):
		$id = $this->detach_caption_target ($caption);
		$caption = $this->phrase ($caption);
	endif;

	// return the result
	if ($caption !== ''):
		return "<figure{$id}>{$file}<figcaption>{$caption}</figcaption></figure>";
	elseif ($id !== ''):
		return "<figure{$id}>{$file}</figure>";
	else:
		return $file;
	endif;
}


/* Parses an Aneamal $string and adds HTML tags at the start and end that
 * correspond to the given Aneamal $mark for gentle ~, heavy * or supplementary
 * emphasis _ or to a quoted " or to a crossed out string before returning it.
 */
private function emphasis (
	string $string,
	string $mark
):	string
{
	if ($mark === '~'):
		return '<i>' . $this->phrase ($string) . '</i>';
	elseif ($mark === '*'):
		return '<b>' . $this->phrase ($string) . '</b>';
	elseif ($mark === '_'):
		return '<u>' . $this->phrase ($string) . '</u>';
	elseif ($mark === '"'):
		return '<q>' . $this->phrase ($string) . '</q>';
	elseif ($mark === '+-'):
		return '<s>' . $this->phrase ($string) . '</s>';
	endif;
}


/* Finds the right mark in a $string that corresponds to the identical left mark
 * whose position in the $string is initially given by &$index. Then &$index is
 * set to the position of the right mark (or to the last character, if no right
 * mark is found) and the part of the $string between the marks is returned. If
 * $no_masking is false, content inside $…$ and |…| (see self::masks for a full
 * list) is not taken into account in the search for the right mark.
 */
private function enclosed (
	string $string,
	int &$index,
	bool $no_masking = false
):	string
{
	// save the mark
	$mark = $string[$index];

	// find the next mark that is not slashed
	$pos = $no_masking? strpos_unslashed ($string, $mark, $index + 1): strpos_unmasked ($string, $mark, self::masks, $index + 1, '`');

	if ($pos === NULL):
		$group = substr ($string, $index + 1);
		$index = strlen ($string) - 1;
		$this->error ('String not closed: expected ' . $mark, 117);
	else:
		$group = substr ($string, $index + 1, $pos - $index - 1);
		$index = $pos;
	endif;

	// remove slashed linebreaks and return the result
	return strip_slashed_breaks ($group);
}


/* Finds the right curly bracket that comes after the left curly bracket whose
 * position in a $string is initially given by &$index. Then &$index is set to
 * the position of the right curly bracket (or to the last character of the
 * string, if no right curly bracket is found) and the part of the $string
 * between the curly brackets is returned unaltered.
 */
private function encurled (
	string $string,
	int &$index
):	string
{
	$pos = strpos_unslashed ($string, '}', $index + 1);

	if ($pos === NULL):
		$group = substr ($string, $index + 1);
		$index = strlen ($string) - 1;
		$this->error ('Curly brackets not closed: expected }', 140);
	else:
		$group = substr ($string, $index + 1, $pos - $index - 1);
		$index = $pos;
	endif;

	return $group;
}


/* Headings marked up with +++ (with rank 2) and + + (with rank 3) hide
 * the contents of their implicitly associated sections by default, but they
 * can be toggled to be visible. This method marks the end of expandable
 * sections of a given or higher numbered, less important $rank, by returning
 * the necessary number of HTML </details> tags.
 */
private function end_expandables (
	int $rank = 1
):	string
{
	$return = '';
	foreach ($this->expandable as $r => $open):
		if ($open and $r >= $rank):
			$return .= '</details>';
			$this->expandable[$r] = false;
		endif;
	endforeach;
	return $return;
}


/* Stores information about an error that can be retrieved with method
 * get_errors.
 */
private function error (
	string $message,
	int $code,
	int $line = NULL
):	void
{
	$this->errors[] = [$message, $code, $line];
}


/* Processes a $block that represents an expandable-section break or
 * expandable-section heading and returns its HTML equivalent. Expandable
 * sections are displayed closed by default. But when a syntax error occurs in a
 * heading, the section is displayed open so that the error message can be found
 * more easily. &$alignment contains class names that apply to an expandable-
 * section heading and need to be set on the HTML <summary> element.
 */
private function expander (
	string $block,
	array &$alignment
):	string
{
	$start = substr ($block, 0, 3);
	$rank = self::expanders[$start];

	// handle expandable-section breaks
	if ($start === $block):
		if ($this->expandable[$rank] === false):
			$this->error ('Expandable-section break without corresponding heading: ' . $start, 166);
		endif;
		if (!empty ($alignment)):
			$this->error ('Alignment for expandable-section break: ' . $start, 214);
		endif;
		$this->sectiontags .= $this->end_expandables ($rank);
		return '';
	endif;

	// end previous expandable sections
	$this->sectiontags .= $this->end_expandables ($rank);

	// find the end of the expandable-section heading
	$endpos = strrpos_unmasked ($block, $start, self::masks, 3);
	if ($endpos === NULL):
		$this->error ('Expandable-section heading not closed: expected ' . $start, 165);
		return '';
	endif;

	// create markup for the new heading
	if (empty ($alignment)):
		$heading = "<summary><h$rank>";
	else:
		$heading = "<summary class='" . implode (' ', $alignment) . "'><h$rank>";
		$alignment = []; // empty so it won't be used in function block anymore
	endif;
	foreach (explode ("\n", trim (strip_slashed_breaks (substr ($block, 3, $endpos - 3)))) as $i => $line):
		$line = $this->phrase ($line);
		if ($i > 0):
			$heading .= "<br><span>$line</span>";
		elseif ($line === ''):
			$this->error ('Expandable-section heading missing', 164);
		else:
			$heading .= $line;
		endif;
	endforeach;
	$heading .= "</h$rank></summary>";
	$this->expandable[$rank] = true;

	// check for unexpected text after the heading, but still in the same block
	if (strlen ($block) !== $endpos + 3):
		$this->error ('Blank line missing after expandable-section heading', 163);
	endif;

	$status = empty ($this->errors)? '': ' open';
	$this->sectiontags .= "<details{$status}>";
	return $heading;
}


/* Processes an Aneamal $block that starts with a left square bracket and
 * contains a corresponding right square bracket positioned at $token_end. This
 * bracketed string could be a file token either followed by a link to a file or
 * by a code block, in which case the block is an embedded file. If there is no
 * link and no code block or the bracketed string does not match the format of a
 * file token (see function parse_file_token in func.php), the block is a simple
 * text paragraph. In either case, the appropriate HTML code is returned.
 */
private function file (
	string $block,
	int $token_end,
	bool $is_list_item = false
):	string
{
	$token = substr ($block, 1, $token_end - 1);

	// not a file, just a bracketed string at the start of a paragraph
	if (
		$token_end === 1
		or $token_end > 2 && $token[1] !== '-' && $token[1] !== ':'
		or !isset ($block[$token_end + 2])
	):
		return $is_list_item? $this->phrase ($block): '<p>' . $this->phrase ($block) . '</p>';
	
	// embedded file
	elseif (substr ($block, $token_end + 1, 2) === "\n|"):
		$lines = explode ("\n", substr ($block, $token_end + 2));
		foreach ($lines as $n => &$line):
			if ($line[0] === '|'):
				$line = substr ($line, 1);
			else:
				return $this->embedded_file (implode ("\n", array_slice ($lines, 0, $n)), $token, implode ("\n", array_slice ($lines, $n)));
			endif;
		endforeach;
		return $this->embedded_file (implode ("\n", $lines), $token);

	// linked file
	elseif (substr ($block, $token_end + 1, 2) === '->'):
		$offset = $token_end + 3;
		if (!isset ($block[$offset]) or ($filename = $this->uri ($block, $offset)) === ''):
			$this->error ('Filename/URI missing: filename expected after ->', 58);
			return '';
		endif;
		return $this->linked_file ($filename, $token, substr ($block, $offset + 1));

	// still not a file, just a bracketed string at the paragraph start
	else:
		return $is_list_item? $this->phrase ($block): '<p>' . $this->phrase ($block) . '</p>';

	endif;
}


/* Loads a linked Aneamal file located at $uri, processes it and returns its
 * content. This is useful for sharing common text between documents or to join
 * multiple parts of a huge document. $tag is the HTML element name which the
 * content gets enclosed in. $tpl specifies an optional Aneamal template name
 * that shall be used with the Aneamal file. Templates can contain metadata,
 * style information in particular.
 */
private function file_aneamal (
	string $uri,
	string $tag,
	string $tpl = NULL
):	string
{
	// prevent an infinite recursion
	static $countdown = self::max_inclusions;
	if (--$countdown < 0):
		$this->error ('Too many Aneamal files included', 66);
		return '';
	endif;

	if (get_uri_type ($uri) !== URI_PATH):
		$this->error ('Forbidden URI type for linked Aneamal file: use a local filename instead', 196);
		return '';
	endif;

	// split path from the query, which can be used to chose only certain lines
	list ($path, $query) = split_uri_tail ($uri);

	// resolve the file path so that it is relative to the Aneamal root
	if (is_null ($canonical = $this->normalize_path ($path))):
		return '';
	endif;

	// read the file content
	if (is_null ($text = file_get_lines ($this->root . $canonical, $query))):
		$this->error ('Aneamal file not readable: ' . $path, 135);
		return '';
	endif;

	// cap the text at the textcap
	$textcap = $this->textcap ?? self::default_textcap;
	if (strlen ($text) > $textcap):
		$this->error ("Aneamal file exceeds textcap of $textcap: " . $canonical, 203);
		$text = substr ($text, 0, $textcap);
	endif;

	// process the file content as Aneamal document
	return $this->process_file_aneamal ($text, dirname ($canonical), basename ($canonical), self::linked, $tag, $tpl);
}


/* Includes the user defined X-module of the given $subtype and passes
 * preprocessed versions of the given $path to it and optionally an $argument.
 * Returns what the module returns, which must be UTF-8 encoded. For example, a
 * module x-audio.php could handle [x-audio]->url markup to integrate an audio
 * file into the document.
 */
private function file_extension (
	string $path,
	string $subtype = NULL,
	string $argument = NULL
):	string
{
	if ($subtype === NULL or $subtype === ''):
		$this->error ('X-module subtype missing', 144);
		return '';
	endif;

	// extend paths or handle data URIs
	$uri_type = get_uri_type ($path);
	if ($uri_type === URI_PATH and !is_null ($canonical = $this->normalize_path ($path))):
		$filename = $this->root . $canonical;
		$href = $this->url . strip_nml_suffix ($canonical);
	elseif ($uri_type === URI_DATA and $this->data_uri ($path) === NULL):
		return '';
	else:
		$filename = NULL;
		$href = $path;
	endif;

	// let the module do its work, check its output, return result
	return $this->use_module ('x-' . $subtype, [
		[
			$path,
			$filename,
			$href,
			$this->root,
			$this->url
		],
		$argument ?? '',
		$this->url . '/aneamal/x-' . $subtype,
		$this->url . $this->dir,
	]);
}


/* Returns an HTML <img> element to include the image located at $path with an
 * optional textual $description for visually impaired readers, bots etc. and
 * an optional textual $hint that is typically displayed as tooltip on mouseover
 * events and comes already prepared for use as a HTML attribute.
 * See https://www.w3.org/TR/html5/semantics-embedded-content.html#alt-text for
 * guidelines on compsoing the $description.
 */
private function file_image (
	string $path,
	string $description = NULL,
	string $hint = NULL
):	string
{
	// resolve the file path
	if (is_null ($source = $this->media_link ($path))):
		return '';
	endif;

	// handle the optional textual alternative
	$alt = isset ($description)? " alt='" . prepare_html_attribute ($description) . "'": '';

	// handle optional hint that is already prepared for use as HTML attribute
	$title = isset ($hint)? " title='$hint'": '';

	// return the image element
	return "<img src='$source'{$alt}{$title}>";
}


/* Returns an HTML <video> element to include the video or audio file located at
 * $path, optionally with a poster image located at $posterpath and text track
 * with closed captions located at $trackpath an optional textual $hint that is
 * typically displayed as tooltip on mouseover events and comes already prepared
 * for use as a HTML attribute. It is intentional to use HTML's <video> element
 * for both audio and video files: HTML's <audio> element does not work with
 * captions or a poster image. While Aneamal also accepts a short description
 * in the same way as for images, i.e. [v:this is what happens in the video]->,
 * that text is not used in the HTML output yet, as there is nothing like
 * the <img> element's alt-attribute for the <video> element in HTML yet.
 */
function file_media (
	string $path,
	string $posterpath = NULL,
	string $trackpath = NULL,
	string $hint = NULL,
	bool $do_loop = false
):	string
{
	$poster = $track = $title = $loop = '';

	// resolve the file path of the media file
	if (is_null ($source = $this->media_link ($path))):
		return '';
	endif;

	// resolve the file path of the poster file
	if ($posterpath !== NULL):
		$posterlink = $this->media_link ($posterpath);
		if ($posterlink !== '' and $posterlink !== NULL):
			$poster = " poster='$posterlink'";
		endif;
	endif;

	// resolve the file path of the track file
	if ($trackpath !== NULL):
		$tracklink = $this->media_link ($trackpath);
		if ($tracklink !== '' and $tracklink !== NULL):
			$track = "<track src='$tracklink' kind='captions' label='CC'>";
		endif;
	endif;

	// handle the optional hint and looping option
	isset ($hint) and $title = " title='$hint'";
	$do_loop and $loop = ' loop';

	return "<video src='$source'{$poster}{$title}{$loop} controls>{$track}<a href='$source'>" . basename ($source) . "</a></video>";
}


/* Loads the image located at $path and creates and saves a usually smaller
 * preview image of it. Returns an HTML <img> element to display the preview
 * with an optional textual $description for visually impaired readers, bots
 * etc. and an optional textual $hint that is typically displayed as tooltip on
 * mouseover events and comes already prepared for use as a HTML attribute.
 * See https://www.w3.org/TR/html5/semantics-embedded-content.html#alt-text for
 * guidelines on compsoing the $description.
 */
private function file_preview (
	string $path,
	string $description = NULL,
	string $hint = NULL
):	string
{
	if (get_uri_type ($path) !== URI_PATH):
		$this->error ('Forbidden URI type for linked image file: use a local filename instead', 201);
		return '';
	endif;

	// resolve the file path to be relative to the Aneamal root directory
	if (is_null ($canonical = $this->normalize_path ($path))):
		return '';
	endif;

	// compose the filename of the original image and check its readability
	$original_filename = $this->root . $canonical;
	if (!is_readable_file ($original_filename)):
		$this->error ('Not a readable file: ' . $path, 127);
		return '';
	endif;

	// get preview image dimensions
	list ($width, $height) = explode_integer_pair (',', $this->pixels ?? self::pixsize);

	// compose a filename for the preview image
	$preview_path = self::pixdir . '/' . md5 ($canonical . ",$width,$height," . filemtime ($original_filename)) . '.jpg';
	$src = encode_special_chars ($this->url . $preview_path);
	$preview_filename = $this->root . $preview_path;

	// prepare image dimensions for existing preview images with fixed size
	if (file_exists ($preview_filename)):
		if ($width > 0 and $height > 0):
			$size = " width='$width' height='$height'";
		else:
			$size = '';
		endif;
	// create the preview image if it does not exist yet
	else:
		if (!is_writable (dirname ($preview_filename))):
			$this->error ('Preview image folder not writable: ' . self::pixdir, 128);
			return '';
		endif;

		$original = @imagecreatefromstring (file_get_contents ($original_filename));
		if ($original === false):
			$this->error ('Image type not supported: ' . $path, 129);
			return '';
		endif;

		// try to read exif data that encodes image orientation
		if ($exif = @exif_read_data ($original_filename, 'IFD0') and isset ($exif['Orientation'])):
			// swap the target width and height for the preview image iff
			// the image orientation indicates a 90° turn to the right or
			// left; we will turn the preview image once it is generated;
			// turning the preview image then is usually more efficient than
			// turning the typically larger original image now
			if ($exif['Orientation'] === 6 or $exif['Orientation'] === 8):
				list ($height, $width) = [$width, $height];
			endif;
		endif;

		// generate the preview image; ditch the original then
		$preview = resize_image ($original, $width, $height);
		imagedestroy ($original);
		if ($preview === NULL):
			$this->error ('Image could not be resized: ' . $path, 130);
			return '';
		endif;

		// rotate preview image according to exif orientation data
		if (isset ($exif['Orientation'])):
			if ($exif['Orientation'] === 3):
				$preview = imagerotate ($preview, 180, 0);
			elseif ($exif['Orientation'] === 6):
				$preview = imagerotate ($preview, 270, 0);
			elseif ($exif['Orientation'] === 8):
				$preview = imagerotate ($preview, 90, 0);
			endif;
		endif;

		$size = " width='" . imagesx ($preview) . "' height='" . imagesy ($preview) . "'";
		imagejpeg ($preview, $preview_filename, self::pixiq);
		imagedestroy ($preview);
	endif;

	// handle the optional textual alternative
	$alt = isset ($description)? " alt='" . prepare_html_attribute ($description) . "'": '';

	// handle optional hint that is already prepared for use as HTML attribute
	$title = isset ($hint)? " title='$hint'": '';

	// return the HTML code for displaying the preview image
	return "<img src='{$src}'{$size}{$alt}{$title}>";
}


/* Returns the content of a text file located at $uri, where $uri refers to a
 * local file or a data URI. Local file paths can be followed by a query to
 * choose only certain lines from the file. If $do_cap is true, the loaded text
 * will be capped if it exceeds a given treshold in order to minimize negative
 * effects when a huge file gets accidentally loaded (e.g. loading an image
 * instead of a text file). If $do_normalize is true, it is checked whether the
 * file is UTF-8 encoded as it should be and otherwise made to be compatible
 * with UTF-8 by stripping invalid characters. It is important to return NULL
 * for missing/unreadable files instead of '' to distinguish this case from
 * empty files. The latter may trigger a certain action in a [t]-module, whereas
 * a [t]-module will not be called in case of a missing/unreadable file.
 */
private function file_raw (
	string $uri,
	bool $do_normalize = true,
	bool $do_cap = true
):	?string
{
	$uri_type = get_uri_type ($uri);

	// split path from the query, which can be used to chose certain lines
	list ($path, $query, $fragment) = split_uri_tail ($uri);

	// extend paths or handle data URIs
	if ($uri_type === URI_PATH):
		if (is_null ($canonical = $this->normalize_path ($path))):
			return NULL;
		elseif (is_null ($text = file_get_lines ($this->root . $canonical, $query))):
			$this->error ('Not a readable file: ' . $path, 192);
			return NULL;
		endif;
	elseif ($uri_type === URI_DATA):
		if (isset ($query) or isset ($fragment)):
			$this->error ('Query or fragment in data URI', 160);
			return NULL;
		elseif (is_null ($text = $this->data_uri ($path, true))):
			return NULL;
		endif;
	else: // prevent remote file inclusion
		$this->error ('Forbidden URI type for file: use local filename or data URI', 200);
		return NULL;
	endif;

	// cap the text at the textcap
	if ($do_cap):
		$textcap = $this->textcap ?? self::default_textcap;
		if (strlen ($text) > $textcap):
			$this->error ("File exceeds textcap of $textcap: " . $path, 193);
			$text = substr ($text, 0, $textcap);
		endif;
	endif;

	return $do_normalize? $this->normalize_utf8 ($text): $text;
}


/* Loads, handles and returns a text file located at $uri and announced with [t]
 * or [t-…]. In the latter case the processing is done by a T-module. Which one
 * is specified by $subtype. The optional $argument is passed to the module. The
 * module could be used to do syntax highlighting or to eval code for example.
 * Its output is expected to be UTF-8 encoded and returned.
 */
private function file_text (
	string $uri,
	string $subtype = NULL,
	string $argument = NULL
):	string
{
	// read the file
	$text = $this->file_raw ($uri, false);

	// process the file content as text
	return is_null ($text)? '': $this->process_file_text ($text, $subtype, $argument);
}


/* Loads a tab-seperated values (TSV) file, which is supposed to be UTF-8
 * encoded and located at $uri. Returns the file contents processed as a HTML
 * <table>. If $is_aneamal is true, Aneamal phrase markup is interpreted in
 * the field names and values. If $transpose is true, the table is transposed,
 * i.e. rows become columns and columns become rows.
 */
private function file_tsv (
	string $uri,
	bool $is_aneamal = false,
	bool $transpose = false
):	string
{
	// read the file and check its encoding
	$data = $this->file_raw ($uri);

	// process the file content as tab-separated values
	return is_null ($data)? '': $this->process_file_tsv ($data, $is_aneamal, $transpose);
}


/* Encodes a $string for use as a HTML id/URL fragment. The returned value only
 * contains small letters, numbers and the hyphen. Returns an empty string on
 * error. Except for the empty string, the same value can only be returned once
 * with $unique being set. Use this to guarantee unique HTML id attributes.
 */
private function fragment_identifier (
	string $string,
	bool $unique = false
):	string
{
	static $uniques = [];

	// create the ID by x_encode'ing the lowercase'd string
	$id = x_encode (trim (str_replace ("\xC2\xA0", ' ', mb_strtolower ($string, 'UTF-8'))));
	if ($id === ''):
		$this->error ('Content missing in target: ' . $string, 19);
		return '';
	endif;

	// check the ID's uniqueness if required
	if ($unique):
		if (in_array ($id, $uniques, true)):
			$this->error ('Target already set: ' . $id, 20);
			return '';
		else:
			$uniques[] = $id;
		endif;
	endif;

	return $id;
}


/* Returns a list of errors that occurrend since this method was last called
 * while processing the current document (or since its start). The reported
 * errors are cleared from the errors array. $message is a heading for the error
 * messages, $lines optionally the array of line numbers (starting at 0) that
 * were processed when the errors occurred. An empty string is returned if no
 * errors occurred.
 */
private function get_errors (
	string $message,
	array $lines = []
):	string
{
	// return empty string if no errors occurred
	if (empty ($this->errors)):
		return '';
	endif;

	$uri = encode_special_chars ($this->errormore ?? self::error_base_url);

	$errors = '';
	$codes = [];
	foreach ($this->errors as [$msg, $code, $line]):
		$line = isset ($line)? 'line ' . strval ($line + 1) . ', ': '';
		$errors .= "<span class='_$code'><br>$code: " . encode_special_chars ($msg) . " ($line<a href='{$uri}{$code}'>more info</a>).</span>";
		$codes[] = $code;
	endforeach;
	$codes = array_unique ($codes, SORT_NUMERIC);
	sort ($codes, SORT_NUMERIC);
	$classes = "_error _" . implode (' _', $codes);

	// compile information about the file/block where the errors occurred
	$info = encode_special_chars ($this->dir . '/' . $this->docname);
	if (!empty ($lines)):
		if (count ($lines) === 1):
			$info .= ', line ' . strval ($lines[0] + 1);
		else:
			$info .= ', lines ' . strval ($lines[0] + 1) . ' to ' . strval (end ($lines) + 1);
		endif;
	endif;

	// compile error message
	$return = "<div class='$classes' style='" . self::error_style . "'><b>$message</b> ($info): $errors</div>";

	// clear error array and return the error messages
	$this->errors = [];
	return $return;
}


/* Finds the end of a word or bare string `...` or URI inside $string whose
 * start is initially given by &$index. Then &$index is set to the position of
 * the last character in the word or bare string or URI, then the word, bare
 * string stripped of its enclosing backticks or URI is returned. If $is_uri is
 * true, less characters are interpreted as markup which would mark the start of
 * a new group and hence end a word, e.g. ~ is a valid character in an URI but
 * not inside a word, as it usually marks a gently emphasized string. Note that
 * &$index will be one less at the end than at the beginning, if no word or bare
 * group or URI is marked by &$index at the beginning. For example consider the
 * string "#. " between the quotation marks: at &$index 1, the dot, no word or
 * bare string or URI is found, so &$index would be set to 0 and an empty string
 * returned.
 */
private function group (
	string $string,
	int &$index,
	bool $is_uri = false
):	string
{
	// handle bare strings separately
	if ($string[$index] === '`'):
		return $this->enclosed ($string, $index, $is_uri);
	endif;

	for ($i = $index, $last = strlen ($string) - 1; $i <= $last; ++$i):
		switch ($string[$i]):
			// a backslash protects the next character
			case '\\':
				$i < $last and ++$i;
				break 1;
			// unprotected whitespace, hint, reference ... end a group
			case "\t":
			case "\n":
			case ' ':
			case '{':
			case '^':
			case '`':
				break 2;
			// punctuation before whitespace, reference or the end ends a group
			case '!':
			case ',':
			case '.':
			case ':':
			case ';':
			case '?':
				if ($i === $last or is_made_of ($string[$i + 1], "\t\n\x20^")):
					break 2;
				else:
					break 1;
				endif;
			// unprotected pairwise markup ends a non-URI group and is not
			// allowed (unprotected) in an URI that is not enclosed by `...`
			case '|':
			case '$':
				if ($is_uri):
					$this->error ('Unprotected special character in URI: ' . $string[$i], 155);
					break 1;
				else:
					break 2;
				endif;
			// unprotected emphasis marks/quotation mark end a non-URI group
			case '*':
			case '_':
			case '~':
			case '"':
				if ($is_uri):
					break 1;
				else:
					break 2;
				endif;
			// metadata and target marks end a non-URI group except at the start
			case '@':
			case '#':
				if ($is_uri or $i === $index):
					break 1;
				else:
					break 2;
				endif;
			// left brackets end a non-URI nesting group
			case '(':
			case '[':
				$j = $i;
				if ($is_uri or is_null ($this->bracketed ($string, $j))):
					break 1;
				else:
					break 2;
				endif;
			// a which is part of strikethrough markup ends a non-uri group
			case '+':
				if (!$is_uri and $i < $last and $string[$i + 1] === '-'):
					break 2;
				else:
					break 1;
				endif;
			// a hyphen which is part of a link arrow ends a group
			case '-':
				if ($i < $last and $string[$i + 1] === '>'):
					break 2;
				else:
					break 1;
				endif;
		endswitch;
	endfor;

	// copy the group from the string and set the index to the group's last
	// character
	$group = substr ($string, $index, $i - $index);
	$index = $i - 1;

	// remove slashed linebreaks and return the result (all linebreaks are
	// slashed linebreaks here)
	return str_replace ("\\\n", '', $group);
}


/* Returns the HTML <head> for this Aneamal document using mainly information
 * from metadata declarations.
 */
private function head (
):	string
{
	// output is always UTF-8 encoded
	$head = "<meta charset='UTF-8'>";

	// the title comes from @title metadata or the main heading
	$head .= '<title>' . encode_special_chars ($this->title . $this->titletail) . '</title>';

	// add general <meta> data, including a meta tag 'generator'
	ksort ($this->metas);
	foreach ($this->metas as $name => $val):
		$head .= sprintf (self::recognized_html_metas[$name], encode_special_chars ($val));
	endforeach;
	$head .= "<meta name='generator' content='Aneamal'>";

	// add <link>s to alternate language files
	ksort ($this->altlangs);
	foreach ($this->altlangs as $lang => $val):
		$head .= "<link rel='alternate' hreflang='$lang' href='" . encode_special_chars ($val) . "'>";
	endforeach;

	// add further meta <link>s
	ksort ($this->links);
	foreach ($this->links as $name => $val):
		$head .= sprintf (self::recognized_html_links[$name], encode_special_chars ($val));
	endforeach;

	// add html meta content
	$head .= implode ($this->metahtmls);

	// add stylesheets
	foreach (array_reverse (array_unique (array_reverse ($this->stylesheets))) as $url):
		$head .= "<link rel='stylesheet' href='" . encode_special_chars ($url) . "'>";
	endforeach;
	foreach (array_reverse (array_unique (array_reverse ($this->metastyles))) as $val):
		$head .= "<style>$val</style>";
	endforeach;

	return $head;
}


/* Returns a HTML link <a> to the given $url. $text is used as link text. If no
 * $text is given, i.e. an empty string, the slightly edited $url is used as
 * link text too. If $unendorsed is true, the link will communicate to search
 * engines that it shall not be seen as a positive sign for the linked location
 * and to browsers that they shall not send a referrer to the linked location.
 */
private function hyperlink (
	string $url,
	string $text = '',
	bool $unendorsed = false
):	string
{
	// prepare URLs depending on their type
	$uri_type = get_uri_type ($url);
	if ($uri_type === URI_PATH):
		if (is_null ($canonical = $this->normalize_path (strip_nml_suffix ($url)))):
			$href = '';
		else:
			$href = $this->url . $canonical;
		endif;
	elseif ($uri_type === URI_FILE):
		$href = $url;
	elseif ($uri_type === URI_DATA):
		if ($this->data_uri ($url) === NULL):
			return '';
		endif;
		$href = $url;
	else:
		$href = $url;
	endif;
	$href = encode_special_chars ($href);

	// generate a description if no description is given
	if ($text === ''):
		if ($uri_type === URI_DATA):
			$this->error ('Data-URI link without description', 118);
			return '';
		elseif ($uri_type === URI_REMOTE):
			$text = $href;
		else:
			$text = encode_special_chars (strip_nml_suffix ($url));
		endif;
	endif;

	// handle optional unendorsement
	$rel = $unendorsed? " rel='nofollow noreferrer'": '';

	// return the URI
	return "<a href='$href'$rel>$text</a>";
}


/* Analyzes the $initial part of a line from a numbered list, which may be an
 * item number like A.2.1. for a new list item. In that case the method returns
 * an array of arrays [0 => numeral type, 1 => decimal value as string] which
 * represent the parts of the item number. Otherwise an empty array is returned.
 * The numeral type is '1', 'A', 'a' or '?' and the decimal value is handled as
 * string so that it works for arbitrarily large values. The parameter
 * $previous must be such an array which represents the previous item's number.
 */
private function item_number (
	string $initial,
	array $previous
):	array
{
	// if the initial part does not end with a dot, it's not an item number
	if ($initial[-1] !== '.'):
		return [];
	endif;

	// split theitem number at the dots, stripping an optional leading hash
	$parts = explode ('.', substr ($initial, $initial[0] === '#'? 1: 0, -1));
	$current = [];

	$maxlevel = count ($parts) - 1;
	$prelevel = count ($previous) - 1;

	// the hierarchy level can only increase by one from item to item
	if ($maxlevel > $prelevel + 1):
		return [];
	endif;

	foreach ($parts as $level => $part):

		// the lack of a numeral in a part means this can not be an item number
		if ($part === '' or $part === '-'):
			return [];
		endif;

		// determine the numerals used for the item number part and its value
		if (is_made_of ($part, '0123456789')):
			$type = '1';
			$value = ltrim ($part, '0') ?: '0';
		elseif (is_made_of ($part, 'ABCDEFGHIJKLMNOPQRSTUVWXYZ')):
			$type = 'A';
			$value = '0';
			foreach (array_map ('ord', str_split (strrev ($part))) as $i => $ascii):
				$value = bcadd ($value, bcmul ((string) ($ascii - 64), bcpow ('26', "$i")));
			endforeach;
		elseif (is_made_of ($part, 'abcdefghijklmnopqrstuvwxyz')):
			$type = 'a';
			$value = '0';
			foreach (array_map ('ord', str_split (strrev ($part))) as $i => $ascii):
				$value = bcadd ($value, bcmul ((string) ($ascii - 96), bcpow ('26', "$i")));
			endforeach;
		elseif (is_made_of ($part, '-', '0123456789')):
			$type = '1';
			$value = '-' . ltrim (substr ($part, 1), '0');
			$value === '-' and $value = '0';
		elseif ($part === '?'): // automatic numbering
			if ($level > $prelevel): // implied: $level === $maxlevel
				$type = '?';
				$value = '1';
			elseif ($level === $maxlevel):
				$type = $previous[$level][0];
				$value = bcadd ($previous[$level][1], '1');
			else:
				$type = $previous[$level][0];
				$value = $previous[$level][1];
			endif;
		else:
			return [];
		endif;

		// the type must be the same as the previous item's on the same level
		if ($level <= $prelevel and $type !== $previous[$level][0]):
			return [];
		endif;

		// the value may not change from the previous item except on maxlevel
		if ($level < $maxlevel and $value !== $previous[$level][1]):
			return [];
		endif;

		$current[] = [$type, $value];

	endforeach;
	return $current;
}


/* Processes the $content of a list item, which can be either a file, a math
 * block or simply text, possibly with phrase markup, and returns the
 * corresponding HTML code.
 */
private function item (
	string $content
):	string
{
	// file candidate, i.e. file or text with bracketed string at its start
	if (isset ($content[0]) and $content[0] === '[' and $offset = strpos_unslashed ($content, ']')):
		return $this->file ($content, $offset, true);
	endif;

	// math block
	if (substr ($content, 0, 2) === '$$'):
		return $this->math_block ($content);
	endif;

	return $this->phrase ($content);
}


/* Identifies the end of a link in $string and returns an HTML <a> element for
 * it. In Aneamal, a link is marked with an arrow -> and &$index must initially
 * provide the location of the greater-than sign in that arrow. The arrow can
 * be either followed by an URI for a regular link or by one of a few special
 * special marks, e.g. @ in which case an URI already declared in metadata will
 * be used. At the end &$index will be set to the position of the last character
 * of the link in $string. The optional $text is used as link text. If the $text
 * is empty, a link text will be generated from what the arrow points at.
 */
private function link (
	string $string,
	int &$index,
	string $text = ''
):	string
{
	// make sure there is a next character
	if (!isset ($string[++$index])):
		$this->error ('Filename/URI missing: filename expected after ->', 39);
		return '';
	endif;

	// distinguish between links to a target in this document ...
	if ($string[$index] === '#'):

		// an empty target implies a jump to the top of the page
		if (!isset ($string[++$index])):
			$group = '';
		else:
			$group = $this->targeted ($string, $index, $text);
		endif;

		// return the HTML link
		if ($group === ''):
			return "<a href='#'>" . ($text === ''? '#': $text) . '</a>';
		else:
			return "<a href='#" . $this->fragment_identifier ($group). "'>$text</a>";
		endif;

	// ... and links declared in metadata ...
	elseif ($string[$index] === '@'):

		// make sure there is another character
		if (!isset ($string[++$index])):
			$this->error ('Metadata name expected after ->@', 92);
			return '';
		endif;

		// retrieve the metadata name
		$name = strip_slashes ($this->group ($string, $index));
		$text !== '' or $text = encode_special_chars ($name);

		// return the last URL defined for the given metadata name
		if (!isset ($this->linkvars[$name])):
			$this->error ('Metadata name after ->@ not declared: ' . $name, 93);
			return '';
		elseif ($this->linkvars[$name] === ''):
			return "<a>$text</a>";
		else:
			return "<a href='" . encode_special_chars ($this->linkvars[$name]) . "'>$text</a>";
		endif;

	// ... and normal link
	else:

		// handle unendorsed links
		if ($unendorsed = $string[$index] === '!'):
			if (!isset ($string[++$index])):
				$this->error ('Filename/URI missing: filename expected after ->!', 94);
				return '';
			endif;
			$address = $this->group ($string, $index, true);
		else:
			$address = $this->uri ($string, $index);
		endif;

		// make sure the URI is not empty and remove escaping backslashes
		if ($address === ''):
			$this->error ('Filename/URI missing: filename expected after ->', 41);
			return '';
		else:
			return $this->hyperlink (strip_slashes ($address), $text, $unendorsed);
		endif;
	endif;
}


/* Prepares a linked file located at $path with the right method for its type,
 * which is encoded in the file $token. See function parse_file_token in
 * func.php for a description of the $token. $extra can contain a caption which
 * must be preceded by white space. For image and media types i, j, v, w it can
 * also contain hints and links before the white space. If a caption is
 * provided or the file is of type i, j, v, w, the processed file and caption
 * are returned as an HTML <figure>, otherwise the result is returned without
 * <figure> element.
 */
private function linked_file (
	string $path,
	string $token,
	string $extra
):	string
{
	list ($type, $subtype, $argument) = parse_file_token ($token);

	if (isset ($argument) and in_array ($type, ['a', 'b', 'd', 'h', 'p', 'q'], true)):
		$this->error ('Unsupported argument for linked file of type: ' . $type, 158);
		return '';
	endif;

	// find optional link, hint, ..., caption and id
	$link = $hint = $poster = $track = NULL;
	$id = $caption = '';
	for ($count = 1; $count <= 4; ++$count):
		if ($extra === '' or $extra === false): // should not be false in PHP7+
			break;
		elseif (is_made_of ($extra[0], "\n\t\x20")):
			$extra = ltrim ($extra);
			$id = $this->detach_caption_target ($extra);
			$caption = $this->phrase ($extra);
			break;
		elseif (in_array ($type, ['i', 'j'], true) and substr ($extra, 0, 2) === '->' and $link === NULL):
			$i = 1;
			$link = $this->link ($extra, $i, '?'); // '?' is one random byte
			$extra = substr ($extra, $i + 1);
		elseif (in_array ($type, ['i', 'j', 'v', 'w'], true) and $extra[0] === '{' and $hint === NULL):
			$i = 0;
			$hint = prepare_html_attribute ($this->encurled ($extra, $i));
			$extra = substr ($extra, $i + 1);
		elseif (in_array ($type, ['v', 'w'], true) and substr ($extra, 0, 2) === '->' and $track === NULL):
			$i = 2;
			if (!isset ($extra[$i])):
				$this->error ('Filename/URI missing: filename expected after ->', 184);
				return '';
 			elseif ($poster === NULL):
 				$poster = strip_slashes ($path);
				$path = $this->uri ($extra, $i);
			else:
				$track = strip_slashes ($this->uri ($extra, $i));
			endif;
			$extra = substr ($extra, $i + 1);
		else:
			$this->error ('Malformed linked-file statement: ' . $extra, 81);
			return '';
		endif;
	endfor;

	// remove backslashes as characters don't need to be protected here
	$path = strip_slashes ($path);

	// call the right method to process the given file type
	if ($type === 'a'):
		$file = $this->file_aneamal ($path, 'div', $subtype);
	elseif ($type === 'b'):
		$file = $this->file_tsv ($path);
	elseif ($type === 'd'):
		$file = $this->file_tsv ($path, true);
	elseif ($type === 'h'):
		$file = $this->file_raw ($path) ?? '';
	elseif ($type === 'i'):
		$file = $this->file_image ($path, $argument, $hint);
	elseif ($type === 'j'):
		$file = $this->file_preview ($path, $argument, $hint);
	elseif ($type === 'p'):
		$file = $this->file_tsv ($path, false, true);
	elseif ($type === 'q'):
		$file = $this->file_tsv ($path, true, true);
	elseif ($type === 't'):
		$file = $this->file_text ($path, $subtype, $argument);
	elseif ($type === 'v'):
		$file = $this->file_media ($path, $poster, $track, $hint);
	elseif ($type === 'w'):
		$file = $this->file_media ($path, $poster, $track, $hint, true);
	elseif ($type === 'x'):
		$file = $this->file_extension ($path, $subtype, $argument);
	else:
		$this->error ('Unrecognized file type: ' . $type, 8);
		return '';
	endif;

	// add a link for images where a link was provided or to the original image
	// for previews without manual link
	if (isset ($link) and $link !== ''):
		$file = substr ($link, 0, -5) . $file . '</a>';
	elseif ($type === 'j' and $file !== ''):
		$file = $this->hyperlink ($path, $file);
	endif;

	// return the optionally captioned result
	if ($caption !== ''):
		return "<figure{$id}>{$file}<figcaption>{$caption}</figcaption></figure>";
	elseif ($id !== ''):
		return "<figure{$id}>{$file}</figure>";
	elseif (in_array ($type, ['i', 'j', 'v', 'w'], true)):
		return "<figure>{$file}</figure>";
	else:
		return $file;
	endif;
}


/* Loads metadata from another Aneamal file, typically an automatically found
 * @meta.nml in the same or an ancestor directory. If an $uri is provided,
 * metadata will be loaded from the given file instead. Returns true, if
 * metadata has been loaded, and false otherwise.
 */
private function load_meta (
	string $uri = NULL
):	bool
{
	$meta = NULL;

	if (isset ($uri)):
		if (get_uri_type ($uri) !== URI_PATH):
			$this->error ('Forbidden URI type for @meta: use a local filename instead', 187);
			return false;
		endif;

		// split path from the query, which can be used to chose certain lines
		list ($path, $query) = split_uri_tail ($uri);

		// resolve the file path so that it is relative to the Aneamal root
		if (is_null ($canonical = $this->normalize_path ($path))):
			return false;
		endif;

		// read the file content
		if (is_null ($text = file_get_lines ($this->root . $canonical, $query))):
			$this->error ('Not a readable file: ' . $path, 202);
			return false;
		endif;

		$meta = new self ($text, dirname ($canonical), $this->url, basename ($canonical), self::meta);
	else:
		foreach (get_directories ($this->dir) as $dir):
			$filename = $dir . '/@meta.nml';
			if (is_readable_file ($this->root . $filename)):
				$meta = new self (file_get_contents ($this->root . $filename), dirname ($filename), $this->url, $filename, self::meta);
				break;
			endif;
		endforeach;
	endif;

	if ($meta === NULL):
		// signal that no metadata has been loaded
		return false;
	endif;

	// @atom
	isset ($meta->links['atom']) and $this->links['atom'] = $meta->links['atom'];
	// @author
	isset ($meta->metas['author']) and $this->metas['author'] = $meta->metas['author'];
	// @class
	$this->classes = $meta->classes;
	// @dir
	$this->direction = $meta->direction;
	// @errormore
	$this->errormore = $meta->errormore;
	// @htmlhead
	$this->metahtmls = $meta->metahtmls;
	// @icon
	isset ($meta->links['icon']) and $this->links['icon'] = $meta->links['icon'];
	// @javascript
	$this->javascripts = $meta->javascripts;
	// @lang|@language
	$this->language = $meta->language;
	// @layout
	$this->layout = $meta->layout;
	// @license
	isset ($meta->links['license']) and $this->links['license'] = $meta->links['license'];
	// @me
	isset ($meta->links['me']) and $this->links['me'] = $meta->links['me'];
	// @pixels
	$this->pixels = $meta->pixels;
	// @publisher
	isset ($meta->metas['publisher']) and $this->metas['publisher'] = $meta->metas['publisher'];
	// @robots
	isset ($meta->metas['robots']) and $this->metas['robots'] = $meta->metas['robots'];
	// @rss
	isset ($meta->links['rss']) and $this->links['rss'] = $meta->links['rss'];
	// @script
	$this->metascripts = $meta->metascripts;
	// @style
	$this->metastyles = $meta->metastyles;
	// @stylesheet
	$this->stylesheets = $meta->stylesheets;
	// @textcap
	$this->textcap = $meta->textcap;
	// @titletail
	$this->titletail = $meta->titletail;
	// @up
	isset ($meta->links['up']) and $this->links['up'] = $meta->links['up'];
	// @viewport
	isset ($meta->metas['viewport']) and $this->metas['viewport'] = $meta->metas['viewport'];

	// copy metadata values
	$this->linkvars = $meta->linkvars;
	$this->textvars = $meta->textvars;
	$this->markvars = $meta->markvars;

	// signal that metadata has been loaded
	return true;
}


/* Applies a function from a math module to a LaTeX math $expression to turn
 * the code into a nice looking formula and returns the result. $is_block says
 * whether the $expression is from a math block, displaystyle in LaTeX
 * terminology, or math phrase markup, textstyle in LaTeX terminology. The math
 * module should be located at aneamal/math/index.php and usually makes use of
 * an external program such as KaTeX, MathJax or mimeTex. The optional $label
 * can contain an equation number for example and a target at the start of the
 * $label can be used to link to the whole math expression.
 */
private function math (
	string $expression,
	bool $is_block = false,
	string $label = ''
):	string
{
	// prepare the HTML wrapper
	if (!$is_block):
		$start = "<span role='math'>";
		$end = '</span>';
	elseif ($label === ''):
		$start = "<div role='math'>";
		$end = '</div>';
	else:
		$id = $this->detach_caption_target ($label);
		$start = "<div{$id} role='math'>";
		$end = " <span class='_label'>" . $this->phrase ($label) . '</span></div>';
	endif;

	$result = $this->use_module ('math', [
		$expression,
		$is_block? 'display': 'text',
		$this->url . '/aneamal/math'
	]);

	if ($result === ''):
		return encode_special_chars ($expression);
	else:
		return $start . $result . $end;
	endif;
}


/* Prepares an Aneamal $block that contains a mathematical LaTeX formula for
 * output. $block must start with two dollar signs and contain two more dollar
 * signs to mark the end of the mathematical expression. It can be followed by
 * an equation label.
 */
private function math_block (
	string $block
):	string
{
	if ($pos = strpos_unslashed ($block, '$$', 2)):
		if (!isset ($block[$pos + 2])):
			return $this->math (substr ($block, 2, $pos - 2), true);
		elseif (!is_made_of ($block[$pos + 2], "\n\t\x20")):
			$this->error ('Missing whitespace between $$ and equation label', 121);
		endif;
		return $this->math (substr ($block, 2, $pos - 2), true, ltrim (substr ($block, $pos + 2)));
	else:
		$this->error ('Math block not closed: expected $$', 6);
		return $this->math (substr ($block, 2), true);
	endif;
}


/* Prepares a given $uri for use as a reference to an image or media file in
 * HTML, for example as a src attribute of an <img> element. Note that an input
 * $uri that starts with a slash is interpreted as relative to the Aneamal root.
 * Returns NULL on error.
 */
function media_link (
	string $uri
):	?string
{
	$uri_type = get_uri_type ($uri);

	// resolve the file path
	if ($uri_type === URI_PATH):
		if (is_null ($canoncial = $this->normalize_path ($uri))):
			return NULL;
		endif;
		$uri = $this->url . $canoncial;
	elseif ($uri_type === URI_FILE):
		$this->error ('Link expected to point to a file: ' . $uri, 204);
		return NULL;
	elseif ($uri_type === URI_DATA):
		if ($this->data_uri ($uri) === NULL):
			return NULL;
		endif;
	endif;

	return encode_special_chars ($uri);
}


/* Handles the metadata declaration of a metadata $name that is recognized by or
 * reserved for special use in Aneamal and whose value $val is plain text (in
 * contrast to a link). Some metadata declarations change the behaviour of the
 * Aneamal-HTML translator (e.g. 'pixels' sets the size for preview images).
 * Other metadata declarations are published in the HTML output as <meta>
 * elements inside the <head>. The output is done by other methods, but this one
 * prepares for it.
 */
private function meta_content (
	string $name,
	string $val
):	void
{
	switch ($name):
		case 'aside':
			if ($val === 'off'):
				$this->aside = '';
			else:
				$this->error ('Invalid @aside value: ' . $val, 182);
			endif;
			return;
		case 'charset':
			$this->encoding = $val;
			return;
		case 'class':
		case 'classes':
			$this->classes = implode (' ', array_map ('trim', explode (',', $val)));
			return;
		case 'dir':
			$dir = strtolower ($val);
			if ($dir === 'ltr' or $dir === 'rtl'):
				$this->direction = $dir;
			else:
				$this->error ('Invalid writing direction declared', 110);
			endif;
			return;
		case 'footer':
			if ($val === 'off'):
				$this->footer = '';
			else:
				$this->error ('Invalid @footer value: ' . $val, 173);
			endif;
			return;
		case 'header':
			if ($val === 'off'):
				$this->header = '';
			else:
				$this->error ('Invalid @header value: ' . $val, 171);
			endif;
			return;
		case 'lang':
		case 'language':
			if (is_made_of ($val, self::alphanumeric . '-')):
				$this->language = $val;
			else:
				$this->error ('Invalid characters in language code', 69);
			endif;
			return;
		case 'layout':
			if ($val === 'manual' or $val === 'blank'):
				$this->layout = $val;
			elseif ($val === 'auto'):
				$this->layout = NULL;
			else:
				$this->error ('Invalid @layout value: ' . $val, 169);
			endif;
			return;
		case 'look':
			if ($val === 'off'):
				$this->look = '';
			else:
				$this->error ('Invalid @look value: ' . $val, 175);
			endif;
			return;
		case 'meta':
			if ($val === 'off'):
				$this->meta = '';
			else:
				$this->error ('Invalid @meta value: ' . $val, 195);
			endif;
			return;
		case 'pixels':
			$this->pixels = $val;
			return;
		case 'role':
			$this->role = $val;
			return;
		case 'textcap':
			$this->textcap = (int) $val;
			return;
		case 'title':
			$this->title = $val;
			return;
		case 'titletail':
			$this->titletail = $val === ''? '': " $val";
			return;
	endswitch;

	if (isset (self::recognized_html_metas[$name])):
		$this->metas[$name] = $val;
	else:
		$this->error ('Reserved metadata name: ' . $name, 190);
	endif;
}


/* Handles the metadata declaration of a metadata $name that is recognized by or
 * reserved for special use in Aneamal and whose value $val is an embedded file.
 * The output is done by other methods, but this one prepares for it.
 */
private function meta_embedded (
	string $name,
	string $val
):	void
{
	switch ($name):
		case 'htmlhead':
			$this->metahtmls[] = $val;
			break;
		case 'script':
			$this->metascripts[] = $val;
			break;
		case 'style':
			$this->metastyles[] = $val;
			break;
		default:
			$this->error ('Embedded file not supported for this metadata name: ' . $name, 154);
	endswitch;
}


/* Handles the metadata declaration of a metadata $name that is recognized by or
 * reserved for special use in Aneamal and whose value $val is an URI (in
 * contrast to plain text). Some metadata declarations change the behaviour of
 * the Aneamal-HTML translator (e.g. 'errormore' sets the base URL for more
 * information in case of errors). Other unescaped variables will later be
 * published in the HTML output as <link> elements inside the <head>. The output
 * is done by other methods, but this one prepares for it.
 */
private function meta_link (
	string $name,
	string $val,  // URI as given in the meta declaration
	string $uri   // URI already checked and prepared for use in links
):	void
{
	switch ($name):
		case 'aside':
			$this->aside = $val;
			return;
		case 'errormore':
			$this->errormore = $uri;
			return;
		case 'footer':
			$this->footer = $val;
			return;
		case 'header':
			$this->header = $val;
			return;
		case 'htmlhead':
			// Loading the file here can be superfluous, because @htmlhead is
			// not used in all kinds of Aneamal files and the kind is not
			// checked at this point. This can be optimized in the future.
			$this->metahtmls[] = $this->file_raw ($val, true, false) ?? '';
			return;
		case 'javascript':
			$this->javascripts[] = $uri;
			return;
		case 'look':
			$this->look = $uri;
			return;
		case 'meta':
			$this->meta = $val;
			return;
		case 'script':
			$this->metascripts[] = $this->file_raw ($val, true, false) ?? '';
			return;
		case 'style':
			$this->metastyles[] = $this->file_raw ($val, true, false) ?? '';
			return;
		case 'stylesheet':
			$this->stylesheets[] = $uri;
			return;
	endswitch;

	if (isset (self::recognized_html_links[$name])):
		$this->links[$name] = $uri;
	elseif (substr ($name, 0, 5) === 'lang-'):
		$lang = substr ($name, 5);
		if ($lang === ''):
			$this->error ('Alternative-language code missing', 85);
		else:
			$this->altlangs[$lang] = $uri;
		endif;
	else:
		$this->error ('Reserved metadata name: ' . $name, 191);
	endif;
}


/* Handles a metadata declaration - that's a metadata value $val assigned to a
 * $name - of various types. Values can be plain text, link URIs or the content
 * of an embedded file. In the latter case, $is_embedded must be true. Metadata
 * can be accessed from within the main body of the document.
 */
private function metadata (
	string $name,
	string $val,
	bool $is_embedded = false
):	void
{
	// check whether the key exists
	if ($name === '' or $name === '\\' or $name === '?'):
		$this->error ('Metadata name missing', 74);
		return;
	endif;

	// check whether the meta data shall overwrite previously set values for
	// the same variable
	if ($name[-1] === '?' and !is_slashed ($name, strlen ($name) - 1)):
		$overwrite = false;
		$name = rtrim (substr ($name, 0, -1));
	else:
		$overwrite = true;
	endif;

	// classify the type of meta variable
	$type = get_meta_type ($name);

	// check the validity of declarations
	if ($type === META_CUSTOM):
		if (strlen ($name) !== 2 or !is_made_of ($name[1], self::alphanumeric)):
			$this->error ('Malformed custom-mark declaration', 95);
			return;
		elseif (in_array ($name, $this->metadecs, true)):
			$this->error ('Custom mark already declared: ' . $name, 96);
			return;
		else:
			// Each custom mark may only be defined locally once.
			// $this->markvars, however, also contains marks defined in a parent
			// document and those may be overwritten. So we track the locally
			// defined metadata names including custom marks in $this->metadecs.
			$this->metadecs[] = $name;
		endif;
	// check the validity of non-custom-mark declarations
	else:
		$name = isset (self::metadata_aliases[$name])? self::metadata_aliases[$name]: strip_slashes ($name);
		// ASCII characters except alphanumerics and hyphen are not allowed
		if (strlen ($name) !== strcspn ($name, ' !"#$%&\'()*+,./:;<=>?@[\]^_`{|}~')):
			$this->error ('Unexpected character in metadata name: ' . $name, 189);
			return;
		elseif (in_array ($name, $this->metadecs, true)):
			$this->error ('Metadata name already declared: ' . $name, 194);
			return;
		elseif (!in_array ($name, self::metadata_multiples)):
			// Each metadata name may only be declared locally once, so we keep
			// track of these.
			$this->metadecs[] = $name;
		endif;
	endif;

	// handle the meta data depending on the form of its content
	if ($is_embedded):

		// handle placeholders
		if ($type === META_CUSTOM):
			if ($overwrite or !isset ($this->markvars[$name])):
				$this->markvars[$name] = $val;
			endif;
			return;
		endif;

		$this->meta_embedded ($name, $val);

	elseif (substr ($val, 0, 2) === '->'): // a link

		$length = strlen ($val);
		$pos = 2;

		// make sure there is an URI
		if ($pos >= $length):
			$val = '';
		elseif (($val = strip_slashes ($this->uri ($val, $pos))) === ''):
			$this->error ('Filename/URI missing: filename expected after ->', 59);
			return;
		endif;

		// check whether the URL is followed by anything, which is forbidden
		if ($length > $pos + 1):
			$this->error ('Characters after link in metadata declaration for @' . $name, 60);
			return;
		endif;

		// handle placeholders
		if ($type === META_CUSTOM):
			if ($overwrite or !isset ($this->markvars[$name])):
				$this->markvars[$name] = $this->file_raw ($val, true, false) ?? '';
			endif;
			return;
		endif;

		// check and normalize URIs
		$uri = $val;
		$uri_type = get_uri_type ($uri);
		if ($uri_type === URI_PATH):
			if (is_null ($val = $this->normalize_path ($val))):
				return;
			endif;
			$uri = $this->url . strip_nml_suffix ($val);
		elseif ($uri_type === URI_DATA and $this->data_uri ($val) === NULL):
			return;
		endif;

		// store the variable for later use in the body
		if ($overwrite or !isset ($this->linkvars[$name])):
			$this->linkvars[$name] = $uri;
		endif;

		// process special meta links further
		if ($type === META_SPECIAL and $val !== ''):
			$this->meta_link ($name, $val, $uri);
		endif;

	else: // simply text

		// identify placeholders mistakenly not defined with a link value
		if ($type === META_CUSTOM):
			$this->error ('Custom mark not declared as link or embedded file', 120);
			return;
		endif;

		$val = strip_slashes ($val);

		// store the variable for later use in the body
		if ($overwrite or !isset ($this->textvars[$name])):
			$this->textvars[$name] = $val;
		endif;

		// process special plain text meta variables further
		if ($type === META_SPECIAL):
			$this->meta_content ($name, $val);
		endif;

	endif;
}


/* Parses a $block that represents options in a multiple-choice question. Each
 * option consists of a key enclosed in curly brackets at the beginning of a
 * line and an answer associated with that key. There can be a $question
 * associated with the options. Returns a HTML <fieldset> with the $question as
 * <legend> and for each option a checkbox for the key and <label> for the
 * answer.
 */
private function multiple_choice (
	string $block,
	string $question = ''
):	string
{
	static $count = 0;
	$choices = [];

	// preprocess each option by extracting the keys and answers
	foreach (explode ("\n{", substr ($block, 1)) as $item):

		// check whether the item is well-formed
		$pos = strpos_unslashed ($item, '}');
		if ($pos === NULL):
			$this->error ('Malformed option: expected }', 112);
			continue;
		elseif (isset ($item[$pos + 1]) and !is_made_of ($item[$pos + 1], "\n\t\x20")):
			$this->error ('Missing whitespace between key and answer in multiple-choice option', 205);
		endif;

		// process the key and the answer
		$key = prepare_html_attribute (substr ($item, 0, $pos));
		$answer = $this->phrase (ltrim (substr ($item, $pos + 1)));

		// count the number of keys to generate a unique ID for each
		++$count;

		// add option
		$choices[] = "<input type='checkbox' id='_$count' value='$key'> <label for='_$count'>$answer</label>";

	endforeach;

	// prepare question and return result
	$question === '' or $question = '<legend>' . $this->phrase ($question) . '</legend>';
	return '<fieldset>' . $question . implode ('<br>', $choices) . '</fieldset>';
}


/* Turns a $path like //foo/./bar/baz/../file into /foo/bar/file by treating a
 * single dot path component as same directory and a double dot component as
 * parent directory. Multiple slashes like // are turned into single ones.
 * Pathes that do not start with a slash like bar/file, i.e. pathes that are
 * relative to the directory of the file that is currently worked on, are
 * changed into pathes with a leading slash that are relative to the Aneamal
 * root directory. The input $path must not be empty and not start with ? or #.
 * Returns null on error, that is when a parent of the Aneamal root directory
 * is referenced.
 */
function normalize_path (
	string $path
):	?string
{
	$parts = [];

	// remove and remember the query string and URL fragment from the path
	$pathlen = strcspn ($path, '?#');
	if ($pathlen < strlen ($path)):
		$tail = substr ($path, $pathlen);
		$path = substr ($path, 0, $pathlen);
	else:
		$tail = '';
	endif;

	// make path relative to the Aneamal root directory if it is not yet
	if ($path[0] !== '/'):
		$path = $this->dir . '/' . $path;
	endif;

	// split the path into its components (folders and file)
	foreach (explode ('/', substr ($path, 1)) as $part):
		// don't add . to the normalized path
		if ($part === '.'):
			continue;
		// drop the current path when encoutering ..
		elseif ($part === '..'):
			if (array_pop ($parts) === NULL):
				$this->error ('Linked parent of Aneamal root directory', 51);
				return NULL;
			endif;
		// add non-empty component to the normalized path
		elseif ($part !== ''):
			array_push ($parts, $part);
		endif;
	endforeach;

	// return the resolved path with leading and possibly trailing slash
	if (empty ($parts)):
		return '/' . $tail;
	elseif ($part === '.' or $part === '..' or $path[-1] === '/'):
		return '/' . implode ('/', $parts) . '/' . $tail;
	else:
		return '/' . implode ('/', $parts) . $tail;
	endif;
}


/* Checks whether a given $string is UTF-8 encoded, normalizes line-breaks to \n
 * and removes the optional initial UTF-8 byte order mark (BOM). Returns the
 * result.
 */
private function normalize_utf8 (
	string $string
):	string
{
	// check wether the string matches UTF-8
	if (!mb_check_encoding ($string, 'UTF-8')):
		$this->error ('Character encoding mismatch: expected UTF-8', 122);
		$string = substitute_nonascii ($string);
	endif;

	// replace Windows- and old Mac-style line breaks with Unix-style ones
	$string = str_replace (["\r\n", "\r"], "\n", $string);

	// remove the optional byte order mark (BOM)
	if (substr ($string, 0, 3) === "\xEF\xBB\xBF"):
		$string = substr ($string, 3);
	endif;

	return $string;
}


/* Parses $lines that together form a numbered list and returns it as HTML <ol>
 * list. Numbered lists may be hierarchical and every sublist may use its own
 * enumeration type. However items on the same level of the same sublist must
 * use the same enumeration type (e.g. decimal numbers).
 */
private function numbered_list (
	array $lines
):	string
{
	$previous = [];
	$list = $item = '';
	$prelevel = $maxlevel = -1;

	foreach ($lines as $line):
		$initial_length = strcspn ($line, "\x20\t");
		$initial = substr ($line, 0, $initial_length);
		$current = $this->item_number ($initial, $previous);

		// if the current line belongs to the latest item
		if ($current === []):
			$item .= "\n" . $line;

		// if the current line constitutes a new item
		else:
			$list .= $this->item ($item);
			$maxlevel = count ($current) - 1;
			$type = $current[$maxlevel][0];
			$value = $current[$maxlevel][1];

			// the item identifier can be a target for links; handle it
			$name = $initial[0] === '#'? $this->fragment_identifier (substr ($initial, 1), true): '';
			$id = $name === ''? '': " id='$name'";

			// if the new item is deeper in the hieararchy than the previous
			if ($maxlevel > $prelevel):
				$list .= $type === '?'? "<ol><li{$id}>": "<ol type='$type'><li{$id} value='$value'>";

			// if the new item is as deep in the hierarchy as the previous
			elseif ($maxlevel === $prelevel):
				$list .= $type === '?'? "</li><li{$id}>": "</li><li{$id} value='$value'>";

			// if the new item is higher in the hierarchy than the previous
			else:
				$list .= str_repeat ('</li></ol>', $prelevel - $maxlevel);
				$list .= $type === '?'? "<li{$id}>": "<li{$id} value='$value'>";
			endif;

			$item = ltrim (substr ($line, $initial_length));
			$previous = $current;
			$prelevel = $maxlevel;
		endif;
	endforeach;

	return $list . $this->item ($item) . str_repeat ('</li></ol>', $maxlevel + 1);
}


/* Parses phrase markup in a $text and returns the text prepared for output as
 * HTML. This method is applied to text inside blocks: to list items, image
 * captions, paragraphs, heading content etc.
 */
private function phrase (
	string $text
):	string
{
	$word = '';
	$output = '';

	// $groupend becomes true without pushing the group into the output, if a
	// group has just ended. This is important so that a following link or hint
	// can be associated with the group or groups before it. If no link or hint
	// follow immediately, the group is pushed to the output before continuing.
	$groupend = false;

	// $linked and $hinted become true when a link or hint are not yet pushed to
	// the output. They can be used to prevent a chain of multiple links or
	// hints which don't make sense.
	$linked = $hinted = false;

	// go through the text, byte by byte
	for ($i = 0, $length = strlen ($text); $i < $length; ++$i):

		$char = $text[$i];

		// link, excluding references to a note
		if ($char === '-' and $length > $i + 1 and $text[$i + 1] === '>'):
			++$i;
			$link = $this->link ($text, $i, $word);
			if ($linked):
				$this->error ('Multiple consecutive links', 88);
			elseif ($link !== ''):
				$word = $link;
			endif;
			$groupend = $linked = true;
			continue;
		// hint
		elseif ($char === '{'):
			$hint = prepare_html_attribute ($this->encurled ($text, $i));
			if ($word === ''):
				$this->error ('Annotated text missing: text expected before {', 56);
			elseif ($hinted):
				$this->error ('Multiple consecutive hints', 89);
			else:
				$word = "<span title='$hint'>$word</span>";
			endif;
			$groupend = $hinted = true;
			continue;
		elseif ($groupend === true):
			$output .= $word;
			$word = '';
			$groupend = $linked = $hinted = false;
		endif;

		switch ($char):
			case '\\': // marks next byte as literal character
				if ($length > ++$i and $text[$i] !== "\n"):
					$word .= encode_printable_ascii ($text[$i]);
				endif;
				break;
			case '&': // custom mark
				if ($length <= ++$i):
					$this->error ('Incomplete custom mark: letter or digit expected after &', 98);
				elseif (isset ($this->markvars['&' . $text[$i]])):
					$word .= $this->markvars['&' . $text[$i]];
				elseif (array_key_exists ($text[$i], self::default_markvars)):
					$word .= self::default_markvars[$text[$i]];
				else:
					$this->error ('Undeclared custom mark: &' . $text[$i], 99);
				endif;
				break;
			case '#': // target
				$output .= $word;
				$word = $this->target ($text, $i);
				$groupend = true;
				break;
			case '@': // metadata
				if ($length <= ++$i):
					$this->error ('Metadata name expected after @', 133);
					$word .= '&#64;';
				elseif (isset ($this->textvars[$name = strip_slashes ($this->group ($text, $i))])):
					$output .= $word;
					$word = encode_special_chars ($this->textvars[$name]);
					$groupend = true;
				else:
					$this->error ('Metadata name after @ not declared: ' . $name, 134);
					$word .= '&#64;' . encode_special_chars ($name);
				endif;
				break;
			case '`': // bare string, e.g. to define a link text
				$output .= $word;
				$word = $this->phrase ($this->enclosed ($text, $i));
				$groupend = true;
				break;
			case '^': // reference to a note
				$output .= $word;
				$word = $this->reference ($text, $i);
				$groupend = $linked = true;
				break;
			case '$': // math
				$output .= $word;
				$word = $this->math ($this->enclosed ($text, $i, true));
				$groupend = true;
				break;
			case '|': // code
				$output .= $word;
				$word = $this->code ($this->enclosed ($text, $i, true));
				$groupend = true;
				break;
			case '}':
				$this->error ('Unmatched right curly bracket: }', 57);
				$word .= '}';
				break;
			case '_': // supplementary emphasis
			case '~': // gentle emphasis
			case '*': // heavy emphasis
			case '"': // quoted string
				$output .= $word;
				$word = $this->emphasis ($this->enclosed ($text, $i), $char);
				$groupend = true;
				break;
			case "\n": // line break
				$output .= $word . '<br>';
				$word = '';
				break;
			case "\t":
			case "\x20": // end of a word
				$output .= $word . "\x20";
				$word = '';
				break;
			case '(':
			case '[': // brackets
				if (is_null ($group = $this->bracketed ($text, $i))):
					$word .= $char;
				else:
					$output .= $word;
					$word = $char . $this->phrase ($group) . self::brackets[$char];
					$groupend = true;
				endif;
				break;
			case '+':
				if ($length > $i + 1 and $text[$i + 1] === '-'):
					$output .= $word;
					$word = $this->emphasis ($this->crossedout ($text, $i), '+-');
					$groupend = true;
				else:
					$word .= '+';
				endif;
				break;
			case '-':
				if ($length > $i + 1 and $text[$i + 1] === '+'):
					$this->error ('Unmatched right cross: -+', 207);
					$word .= '-+';
					++$i;
				else:
					$word .= '-';
				endif;
				break;
			case '<':
				$word .= '&#60;';
				break;
			case '>':
				$word .= '&#62;';
				break;
			default:
				$word .= $char;
		endswitch;

	endfor;

	return $output . $word;
}


/* Removes comments (i.e. lines starting with %) and interprets and removes
 * metadata declarations (i.e. lines starting with @). Despite removing lines,
 * the array keys of $this->lines are preserved. Charset metadata is ignored,
 * iff $ignore_charset is true.
 * There are two kinds of metadata, regular and embedded file metadata. The
 * initial line of regular metadata must contain an unslashed colon separating
 * the name from the value. The value may span multiple lines where all lines
 * start with @ and all but the last line end with \ and there are no other
 * lines in between except for comments. The initial line of embedded file
 * metadata does not contain an unslashed colon. Its name is the whole initial
 * line (without @ and trimmed). Its value is the embedded file content defined
 * by the following lines that start with @ followed by |.
 */
private function preprocess_comments_meta (
):	void
{
	$state = 0; // 0 (none), 1 (regular) or 2 (embedded)
	$val = '';
	foreach (array_keys ($this->lines) as $n):

		// get the leftmost byte
		$left = $this->lines[$n][0] ?? NULL;

		// check for reserved semicolon
		if ($left === ';'):
			$this->error ('Reserved first character in line: semicolon ;', 210, $n);
		endif;

		// if in a comment line ...
		if ($left === '%'):

			// ... remove the line
			unset ($this->lines[$n]);

		// if in a metadata line ...
		elseif ($left === '@'):

			// get the line without the initial @
			$line = ltrim (substr ($this->lines[$n], 1));

			// if in an embedded file already
			if ($state === 2):
				if (isset ($line[0]) and $line[0] === '|'):
					if ($val === ''):
						$val = substr ($line, 1);
					else:
						$val .= "\n" . substr ($line, 1);
					endif;
				else:
					$this->metadata ($name, $val, true);
					$state = 0;
				endif;
			endif;

			// if not in multiline metadata yet
			if ($state === 0):
				$statement = explode_unslashed (':', $line, 2);
				if (isset ($statement[1])):
					$name = rtrim ($statement[0]);
					$val = ltrim ($statement[1]);
					$state = 1;
				elseif ($statement[0] !== ''):
					$name = $statement[0];
					$val = '';
					$state = 2;
				endif;
			// if in multiline metadata already
			elseif ($state === 1):
				$val = substr ($val, 0, -1) . $line;
			endif;

			// if regular metadata doesn't continue as multiline metadata
			if ($state === 1 and ($val === '' or $val[-1] !== '\\' or is_slashed ($val, strlen ($val) - 1))):
				$this->metadata ($name, $val);
				$state = 0;
			endif;

			unset ($this->lines[$n]);

		// if not in a metadata line, but an embedded file had been open
		elseif ($state === 2):

			$this->metadata ($name, $val, true);
			$state = 0;

		// if not in a metadata line, but a metadata line was expected
		elseif ($state === 1):

			$this->error ('Malformed multiline metadata declaration for @' . $name, 100);
			$this->metadata ($name, substr ($val, 0, -1));
			$state = 0;

		endif;

	endforeach;

	// if embedded file/multiline metadata is still open at the document end
	if ($state === 2):
		$this->metadata ($name, $val, true);
	elseif ($state === 1):
		$this->metadata ($name, substr ($val, 0, -1));
	endif;
}


/* Removes white space from the beginning and end of each line in the file and
 * converts sandwich markup to line style markup. Sandwich style markup begins
 * with a line like "/prefix/until" and ends with a line like "until" (quotes
 * not included). Converting it to line style markup means removing the initial
 * and final lines and prepending "prefix" to every line in between. Despite
 * removing lines, the array keys of $this->lines are preserved.
 */
private function preprocess_lines (
):	void
{
	$until = NULL;
	foreach (array_keys ($this->lines) as $n):

		// if not inside a sandwich yet
		if ($until === NULL):
			$this->lines[$n] = trim ($this->lines[$n]);

			// if in the initial line of sandwich markup
			if (isset ($this->lines[$n][0]) and $this->lines[$n][0] === '/'):
				$pos = strpos_unslashed ($this->lines[$n], '/', 1);
				if ($pos === NULL):
					$this->error ('Invalid sandwich markup: expected second /', 141, $n);
				else:
					$prefix = ltrim (strip_slashes (substr ($this->lines[$n], 1, $pos - 1)));
					if ($prefix === ''):
						$this->error ('Prefix missing in sandwich markup', 143, $n);
					else:
						$until = ltrim (substr ($this->lines[$n], $pos + 1));
						unset ($this->lines[$n]);
					endif;
				endif;
			endif;

		// if in the final line of sandwich markup
		elseif (trim ($this->lines[$n]) === $until):
			$until = NULL;
			unset ($this->lines[$n]);

		// if inside a sandwich
		else:
			$this->lines[$n] = trim ($prefix . $this->lines[$n]);
		endif;

	endforeach;

	// if sandwich markup is still open at the document end
	if ($until !== NULL and $until !== ''):
		$this->error ('Bottom sandwich delimiter missing: ' . $until, 142);
	endif;
}


/* Parses an Aneamal text that has been included as linked or embedded file or
 * as a block quotation. A template which adds standard styles and scripts may
 * be used. These and the file's stylesheets and scripts are added to this
 * file's corresponding array. The file's meta title and general metadata are
 * not used.
 *
 * $text:       Aneamal text
 * $dir:        directory in which the parsed file is located relative to the
 *              Aneamal root, starting with a slash; needed to locate other
 *              files referenced from this
 * $docname:    identifier for this document in error messages
 * $kind:       an integer corresponding to the kind of Aneamal document, either
 *              self::main, self::embedded, self::linked, self::template,
 *              self::layout or self::meta
 * $tag:        HTML tag name to enclose the document content
 * $tpl:        name of a metadata template file
 * $citation:   optionally a string that references the source of the text, this
 *              makes sense for quotation blocks
 */
private function process_file_aneamal (
	string $text,
	string $dir,
	string $docname,
	int $kind,
	string $tag,
	string $tpl = NULL,
	string $citation = ''
):	string
{
	// cache for templates: they only need to be loaded once for multiple uses
	static $templates = [];

	$role = '';
	$class = '';
	$pixels = NULL;
	$linkvars = $this->linkvars;
	$textvars = $this->textvars;
	$markvars = $this->markvars;

	// handle the optional template
	if (isset ($tpl)):
		if (!isset ($templates[$tpl])):
			if (
				is_readable_file ($this->root . ($tplfile = '/aneamal/a-' . $tpl . '/index.nml'))
				or is_readable_file ($this->root . ($tplfile = '/aneamal/a-' . $tpl . '.nml'))
			):
				$templates[$tpl] = new self (
					file_get_contents ($this->root . $tplfile),
					dirname ($tplfile),
					$this->url,
					$tplfile,
					self::template,
					$linkvars,
					$textvars,
					$markvars
				);
				$this->javascripts = array_merge ($this->javascripts, $templates[$tpl]->javascripts);
				$this->metascripts = array_merge ($this->metascripts, $templates[$tpl]->metascripts);
				$this->stylesheets = array_merge ($this->stylesheets, $templates[$tpl]->stylesheets);
				$this->metastyles = array_merge ($this->metastyles, $templates[$tpl]->metastyles);
				if (!empty ($templates[$tpl]->lines) and implode ($templates[$tpl]->lines) !== ''):
					$this->error ('Content in template ' . $tplfile, 151);
				endif;
			else:
				$this->error ('Template not readable: a-' . $tpl, 152);
			endif;
		endif;
		if (isset ($templates[$tpl])):
			$role = $templates[$tpl]->role;
			$class = $templates[$tpl]->classes;
			$pixels = $templates[$tpl]->pixels;
			$linkvars = $templates[$tpl]->linkvars;
			$textvars = $templates[$tpl]->textvars;
			$markvars = $templates[$tpl]->markvars;
		endif;
	endif;

	$doc = new self ($text, $dir, $this->url, $docname, $kind, $linkvars, $textvars, $markvars);

	// pass preview image size settings from template or main file to the $doc
	$doc->pixels = $doc->pixels ?? $pixels ?? $this->pixels;

	// pass base URI for detailed error explanations to the $doc
	is_null ($doc->errormore) and $doc->errormore = $this->errormore;

	// pass size limit for text file inclusions to the $doc
	is_null ($doc->textcap) and $doc->textcap = $this->textcap;

	// pass language data to the $doc iff it lacks it
	is_null ($doc->language) and $doc->language = $this->language;
	is_null ($doc->direction) and $doc->direction = $this->direction;

	// generate the document body of the file
	$body = $doc->body ();
	if ($body === ''):
		return '';
	endif;

	// inherit styles and scripts
	$this->javascripts = array_merge ($this->javascripts, $doc->javascripts);
	$this->metascripts = array_merge ($this->metascripts, $doc->metascripts);
	$this->stylesheets = array_merge ($this->stylesheets, $doc->stylesheets);
	$this->metastyles = array_merge ($this->metastyles, $doc->metastyles);

	// prepare $doc's language data iff different from this document's data
	$lang = $dir = '';
	if ($this->language !== $doc->language):
		$lang = " lang='" . $doc->language . "'";
	endif;
	if ($this->direction !== $doc->direction):
		$dir = " dir='" . $doc->direction . "'";
	endif;

	// prepare role and class, combined from the template and the $doc
	$role = trim ($role . ' ' . $doc->role);
	if ($role !== ''):
		$role = " role='" . encode_special_chars ($role) . "'";
	endif;
	$class = trim ($class . ' ' . $doc->classes);
	if ($class !== ''):
		$class = " class='" . encode_special_chars ($class) . "'";
	endif;

	// handle the optional citation and id
	$id = '';
	if ($citation !== ''):
		$id = $this->detach_caption_target ($citation);
		$citation = '<cite>' . $this->phrase ($citation) . '</cite>';
	endif;

	return "<{$tag}{$id}{$dir}{$lang}{$role}{$class}>{$body}{$citation}</$tag>";
}


/* Lets a T-module specified by $subtype process a give $text. The T-module is
 * provided with an optional $argument and its returned result is expected to be
 * UTF-8 encoded. If no $subtype is provided, the $text is handled as
 * preformatted plain text. The result is returned.
 */
function process_file_text (
	string $text,
	string $subtype = NULL,
	string $argument = NULL
):	string
{
	// handle text which is not processed by a module
	if ($subtype === NULL):
		return "<pre class='_plain'>" . encode_special_chars ($this->normalize_utf8 ($text)) . '</pre>';
	endif;

	// let the module do its work
	return $this->use_module ('t-' . $subtype, [
		$text,
		$argument ?? '',
		$this->url . '/aneamal/t-' . $subtype,
		$this->url . $this->dir
	]);
}


/* Translates tab-seperated values ($tsv) into an HTML table and returns it.
 * Parameters allow to $parse_phrase_markup within values and to transpose the
 * whole table.
 */
private function process_file_tsv (
	string $tsv,
	bool $parse_phrase_markup = false,
	bool $transpose_table = false
):	string
{
	// remove optional line break from the end
	if ($tsv !== '' and $tsv[-1] === "\n"):
		$tsv = substr ($tsv, 0, -1);
	endif;

	// split the file into lines and fields
	$data = array_map (function ($line) {return explode ("\t", $line);}, explode ("\n", $tsv));

	// check the number of fields per record and parse values
	$columns = 0;
	$tag = 'th';
	foreach ($data as $row => $record):
		foreach ($record as $col => $field):
			$data[$row][$col] = "<$tag>" . ($parse_phrase_markup? $this->phrase ($field): encode_special_chars ($field)) . "</$tag>";
		endforeach;
		$count = count ($record);
		if ($columns === 0):
			$columns = $count;
			$tag = 'td';
		elseif ($count < $columns):
			$data[$row] = array_pad ($data[$row], $columns, '<td></td>');
		elseif ($count > $columns):
			$this->error ('Too many fields in TSV line ' . strval ($row + 1), 183);
			return '';
		endif;
	endforeach;

	// optionally transpose data
	if ($transpose_table):
		$ndata = array ();
		foreach ($data as $row => $record):
			foreach ($record as $col => $field):
				$ndata[$col][$row] = $field;
			endforeach;
		endforeach;
		$data = $ndata;
	endif;

	// finish and return the table
	foreach ($data as $row => $record):
		$data[$row] = '<tr>' . implode ($record) . '</tr>';
	endforeach;
	return '<table>' . implode ($data) . '</table>';
}


/* Identifies the end of a reference to a note in $string and translates the
 * reference to HTML, i.e. returns a <sup>erscript HTML <a> element. Initially
 * &$index must give the position of a character ^ in $string which marks the
 * reference and is followed by the name of a target on the same page, usually
 * in a note. At the end of the method, &$index will be set to the position of
 * the last character of the reference.
 */
private function reference (
	string $string,
	int &$index
):	string
{
	// make sure there is a next character
	if (!isset ($string[++$index])):
		$this->error ('Target text missing in reference: target expected after ^', 71);
		return '';
	endif;

	// extract and process target name
	$text = '';
	$name = $this->fragment_identifier ($this->targeted ($string, $index, $text));

	// return the HTML output
	if ($name === ''):
		return "<sup>$text</sup>";
	else:
		return "<sup><a href='#$name'>$text</a></sup>";
	endif;
}


/* Sets non-static properties of this nml2html object to their initial value.
 */
private function reset (
):	void
{
	foreach (get_class_vars (get_class ($this)) as $property => $initial):
		$this->$property = $initial;
	endforeach;
}


/* Processes a $block that represents a section break or heading with optional
 * sublines (excluding expandable-section breaks and headings, which are
 * processed by the method expander) and returns their HTML equivalent.
 */
private function sectioner (
	string $block
):	string
{
	$start = substr ($block, 0, 3);
	$rank = self::sectioners[$start];

	// end previous expandable sections
	$this->sectiontags .= $this->end_expandables ($rank);

	// handle section breaks
	if ($start === $block):
		if ($rank === 1):
			$this->error ('Main heading incomplete', 188);
			return '';
		else:
			return "<hr class='_h{$rank}'>";
		endif;
	endif;

	// find the end of the heading
	$endpos = strrpos_unmasked ($block, $start, self::masks, 3);
	if ($endpos === NULL):
		$this->error ('Heading not closed: expected ' . $start, 4);
		return '';
	endif;

	// translate the heading to HTML
	$return = "<h$rank>";
	$lines = explode ("\n", $raw = trim (strip_slashed_breaks (substr ($block, 3, $endpos - 3))));
	foreach ($lines as $i => $line):
		$line = $this->phrase ($line);
		if ($i > 0):
			$return .= "<br><span>$line</span>";
		elseif ($line === ''):
			$this->error ('Heading text missing', 176);
		else:
			$return .= $line;
		endif;
	endforeach;
	$return .= "</h$rank>";

	// use the main heading as <title> if no @title was declared as metadata
	if ($rank === 1 and $this->title === ''):
		$this->title = str_replace ("\n", "\x20", strip_slashes (ltrim ($raw)));
	endif;

	// return the heading with the optional sublines
	if (strlen ($block) === $endpos + 3):
		return $return;
	elseif ($block[$endpos + 3] !== "\n"):
		$this->error ('Line feed missing after heading', 91);
		return $return;
	else:
		$sublines = $this->phrase (ltrim (substr ($block, $endpos + 3)));
		return "<div role='group'>$return<p>$sublines</p></div>";
	endif;
}


/* Parses a $block which represents a tagged list or double-tagged list. Returns
 * a HTML description list (<dl>) for a tagged list or a HTML <table> for a
 * double-tagged list. In tagged lists the same tag may occur in multiple items,
 * whereas the same tag-pair may only occur once in a double-tagged list.
 */
private function tagged_list (
	string $block
):	string
{
	$items = [];
	$dimension = 0;
	$operator = NULL;

	// this foreach preprocesses each item by extracting the tags and content
	foreach (explode ("\n<", substr ($block, 1)) as $item):

		$close = -1;
		$start = 0;
		$tags = [];
		$count = 0;
		$empty = 0;

		// save the item's tags, identified by pointy brackets < >
		while ($close = strpos_unmasked ($item, '>', self::masks, $close + 1) or $close === 0):
			// continue searching if > is preceded by an unslashed -, which
			// means it defines a link, not the end of the tag
			if ($close > 0 and $item[$close - 1] === '-' and !is_slashed ($item, $close - 1)):
				continue;
			endif;
			// save the tag, count empty and total tags
			$tags[$count] = trim (substr ($item, $start, $close - $start));
			$tags[$count] === '' and ++$empty;
			++$count;
			// check whether > is followed by <, which means another tag follows
			// for the same item, otherwise stop searching
			if (isset ($item[$close + 1]) and $item[$close + 1] === '<'):
				$start = $close + 2;
			else:
				break;
			endif;
		endwhile;

		// the first item's number of tags determines the list's dimension
		if ($dimension === 0):
			$dimension = $count;
		endif;

		// save the item if a closing > was found, the number of tags is
		// consistent with the list's dimension and there are no empty tags or
		// all tags are empty in a list with dimension greater than one
		if ($close === NULL):
			$this->error ('Malformed tagged-list item: expected >', 7);
		elseif ($count !== $dimension):
			$this->error ("Inconsistent number of tags: $count instead of $dimension", 14);
		elseif ($empty > 0 and $dimension > 1 and $empty === $dimension):
			if ($operator === NULL):
				$operator = ltrim (substr ($item, $close + 1));
			else:
				$this->error ('Origin already set', 105);
			endif;
		elseif ($empty > 0):
			$this->error ('Empty tag in tagged list', 106);
		else:
			$items[] = array_merge ($tags, [ltrim (substr ($item, $close + 1))]);
		endif;

	endforeach;

	// if dimension is 1, render the list as HTML description list …
	if ($dimension === 1):

		$list = '';
		foreach ($items as list ($tag, $content)):
			$list .= '<dt>' . $this->phrase ($tag) . '</dt><dd>' . $this->item ($content) . '</dd>';
		endforeach;
		return "<dl>$list</dl>";

	// … or as HTML table if the dimension is 2
	elseif ($dimension === 2):

		$array = $x_axis = $y_axis = [];

		// fill a two-dimensional $array with the content of the items and store
		// the tags as column-/row headings for the x- and y-axes
		foreach ($items as list ($y, $x, $content)):
			in_array ($y, $y_axis, true) or $y_axis[] = $y;
			in_array ($x, $x_axis, true) or $x_axis[] = $x;
			// check whether the same tag-pair was previously used
			if (isset ($array[$y][$x])):
				$this->error ("Tag pair already used: <$y><$x>", 15);
			else:
				$array[$y][$x] = $content;
			endif;
		endforeach;

		// build the first row of the table
		$table = '<tr><td>' . $this->item ($operator ?? '') . '</td>';
		foreach ($x_axis as $x):
			$table .= '<th>' . $this->phrase ($x) . '</th>';
		endforeach;
		$table .= '</tr>';

		// build the body of the table
		foreach ($y_axis as $y):
			$table .= '<tr><th>' . $this->phrase ($y) . '</th>';
			foreach ($x_axis as $x):
				$table .= '<td>' . $this->item ($array[$y][$x] ?? '') . '</td>';
			endforeach;
			$table .= '</tr>';
		endforeach;

		return "<table>$table</table>";

	// send an error for tagged lists with dimension greater than two
	else:

		$this->error ("Too many tags: $dimension", 49);
		return '';

	endif;
}


/* Identifies the end of a target in a $string, translates it into an HTML
 * <span> element with id attribute and returns it. The position of the first
 * character # of the target must be given in &$index at the beginning and is
 * set to the position of the last character at the end. If the # is directly
 * followed by a hint {…}, the target will be invisible in the HTML output, but
 * it will still be possible to link to it. If the # is directly followed by
 * metadata @, the metadata value will be displayed in the HTML output, but the
 * metadata name must be used to link to it.
 */
private function target (
	string $string,
	int &$index
):	string
{
	// make sure there is a next character
	if (!isset ($string[++$index])):
		$this->error ('Target text missing: text expected after #', 42);
		return '';
	endif;

	// extract and process the target name
	$text = '';
	if ($string[$index] === '{'):
		$group = $this->encurled ($string, $index);
		$name = $this->fragment_identifier ($group, true);
		$hint = " title='" . prepare_html_attribute ($group) . "'";
	else:
		$name = $this->fragment_identifier ($this->targeted ($string, $index, $text), true);
		$hint = '';
	endif;

	// prepare and return the HTML output
	$name === '' or $name = " id='$name'";
	return "<span{$name}{$hint}>$text</span>";
}


/* Identifies the end of a target name in a $string, either where the target is
 * set (except invisible targets #{...}), or where it is referenced/linked to.
 * &$index must give the position of the first character of the target's name
 * initially, excluding target/reference/link marks #, ^, ->, and is set to its
 * last character's position at the end. If an empty string is passed as $text
 * parameter, that variable will be replaced with parsed text generated from the
 * target from the target name for display. Returns the target name.
 */
private function targeted (
	string $string,
	int &$index,
	string &$text = NULL
):	string
{
	$char = $string[$index];

	if (in_array ($char, ['*', '~', '_', '"'], true)):
		$group = $this->enclosed ($string, $index);
		if ($text === ''):
			$text = $this->emphasis ($group, $char);
		endif;
	elseif ($char === '|'):
		$group = $this->enclosed ($string, $index, true);
		if ($text === ''):
			$text = $this->code ($group);
		endif;
	elseif ($char === '$'):
		$group = $this->enclosed ($string, $index, true);
		if ($text === ''):
			$text = $this->math ($group);
		endif;
	elseif (array_key_exists ($char, self::brackets) and !is_null ($group = $this->bracketed ($string, $index))):
		if ($text === ''):
			$text = $char . $this->phrase ($group) . self::brackets[$char];
		endif;
	elseif ($char === '+' and isset ($string[$index + 1]) and $string[$index + 1] === '-'):
		$group = $this->crossedout ($string, $index);
		if ($text === ''):
			$text = $this->emphasis ($group, '+-');
		endif;
	else:
		$group = $this->group ($string, $index);
		if ($text === ''):
			$text = $this->phrase ($group);
		endif;
	endif;

	return $group;
}


/* Identifies the end of an URI in a given $string whose first character's
 * position is initially given by &$index, sets &$index to the position of the
 * last character and returns the URI. If the URI starts with a comma, it is
 * expanded into a full data URI of base64-encoded UTF-8 encoded text. Other
 * link qualifiers like @, !, # are not allowed. Note that $&index will be one
 * less at the end than at the beginning if no URI is found at the given
 * position.
 */
private function uri (
	string $string,
	int &$index
):	string
{
	if (is_made_of ($string[$index], '#!@')):
		$this->error ('Invalid link modifier: ' . $string[$index] . ' not allowed after -> in this case', 107);
		return $this->group ($string, $index, true);
	elseif ($string[$index] !== ','):
		return $this->group ($string, $index, true);
	elseif (isset ($string[++$index])):
		return 'data:;charset=UTF-8;base64,' . $this->group ($string, $index, true);
	else:
		return 'data:;charset=UTF-8;base64,';
	endif;
}


/* Executes a $module, supplying it with $arguments, and returns its result,
 * which is expected to be UTF-8 encoded. This method is used for T-modules
 * [t-...], X-module [x-...] and the math module which parses $$...$$ and $...$.
 * The composition of the $arguments array depends on the kind of module.
 */
private function use_module (
	string $module,
	array $arguments
):	string
{
	$path = __DIR__ . '/' . $module;
	try {
		$result = execute_module ($path . '/index.php', $arguments) ?? execute_module ($path . '.php', $arguments);
	} catch (\ParseError $e) {
		$this->error ("ParseError in module $module: \"" . $e->getMessage () . '" in ' . strip_root ($e->getFile ()) . ', line ' . $e->getLine (), 198);
		return '';
	} catch (\TypeError $e) {
		$this->error ("TypeError in module $module: " . strip_root ($e->getFile ()) . ', line ' . $e->getLine (), 199);
		return '';
	} catch (\Throwable $e) {
		$this->error ("Error in module $module: \"" . $e->getMessage () . '" in ' . strip_root ($e->getFile ()) . ', line ' . $e->getLine (), 197);
		return '';
	}

	if ($result === NULL):
		$this->error ('Module not found: ' . $module, 186);
		return '';
	endif;

	// check text encoding and return the result
	return $this->normalize_utf8 ($result);
}


} // class nml2html


// Version: 27.1