<?php // Please do not edit this file. If you do, save as UTF-8 without BOM!

/* Copyright 2010-2020 Martin Janecke <aneamal@prlbr.de>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


// declare (strict_types = 1); // only during development

namespace prlbr\aneamal;

// error_reporting (E_ALL); // only during development


// filename
$query = $_SERVER['QUERY_STRING'];
$file  = __DIR__ . '/..' . $query;
$base  = basename ($query);


// The restriction on .nml prevents files with sensitive data to become publicly
// available, for example data in .htaccess or .php files.
if (substr ($query, -4) === '.nml' and is_readable ($file)):

	// Aneamal file names with an initial @-symbol are for automatic inclusion
	// in other responses and not made to be accessed directly, hence the 403
	// Forbidden status. They are nevertheless processed normally when accessed
	// directly, which is useful for trouble shooting.
	if ($forbidden = $base[0] === '@'):
		header ('HTTP/1.1 403 Forbidden');
	endif;

	header ('Content-Type: text/html; charset=UTF-8');

	// At most one of these cache controls will be changed to true.
	$feedcache = $readcache = $dropcache = false;

	// AneamalCache "on" means use cache except for HTTP POST requests; "forced"
	// means use cache independent of HTTP method; a number is like "on" except
	// that cached files may not be older than the given number of seconds.
	if ($cacherule = getenv ('AneamalCache')):
		$cacherule = strtolower ($cacherule);
		$cache = __DIR__ . '/cache/' . md5 ($query) . '.html';
		if ($_SERVER['REQUEST_METHOD'] === 'POST' and $cacherule !== 'forced'):
			$dropcache = is_readable ($cache);
		elseif (is_readable ($cache)):
			$cachetime = filemtime ($cache);
			if ($cacherule === 'on' or $cacherule === 'forced'):
				$readcache = $cachetime > filemtime ($file);
				$feedcache = !$readcache;
			elseif ($cachelife = (int) $cacherule):
				$readcache = $cachetime > filemtime ($file) && $cachetime + $cachelife > time ();
				$feedcache = !$readcache;
			endif;
		elseif ($cacherule === 'on' or $cacherule === 'forced' or (int) $cacherule):
			$feedcache = true;
		endif;
	endif;

	// write the document to the output
	if ($readcache):

		// Implements https://tools.ietf.org/html/rfc7232#section-2.2
		header ('Last-Modified: ' . gmdate ('D, d M Y H:i:s', $cachetime) . ' GMT');

		// Implements https://tools.ietf.org/html/rfc7232#section-3.3
		if (
			isset ($_SERVER['HTTP_IF_MODIFIED_SINCE'])
			and $_SERVER['REQUEST_METHOD'] === 'GET' || $_SERVER['REQUEST_METHOD'] === 'HEAD'
			and $remotetime = @strtotime ($_SERVER['HTTP_IF_MODIFIED_SINCE'])
			and $remotetime >= $cachetime
			and !$forbidden
		):
			header ('HTTP/1.1 304 Not Modified');
		else:
			readfile ($cache);
		endif;

	else:

		// start timing of document generation
		$start = microtime (true);

		// include the class file which does the Aneamal-HTML-translation
		require __DIR__ . '/html.php';

		// create an object doing the Aneamal-HTML-translation and pass the
		// Aneamal file contents to it
		$converter = new nml2html (
			file_get_contents ($file),
			dirname ($query),
			substr ($_SERVER['SCRIPT_NAME'], 0, -17), // 17: /aneamal/main.php
			$base
		);
		$html = $converter->document ();

		// stop timing
		$duration = microtime (true) - $start;

		// print the HTML document
		print $html . '<!--' . strval ($duration) . '-->';

		// write the file to the cache
		if ($feedcache):
			if (!is_writable ('cache')):
				print '<!--cache-not-writable-->';
			elseif (@file_put_contents ($cache, $html . '<!--' . md5 ($query) . '-from-cache-->')):
				print '<!--written-to-cache-->';
			else:
				print '<!--failed-to-cache-->';
			endif;
		elseif ($dropcache):
			touch ($cache, 0);
		endif;

	endif;

// ... or return HTTP status 404 if the file isn't a readable Aneamal-file
else:

	header ('HTTP/1.1 404 Not Found');
	header ('Content-Type: text/html');
	print '<!doctype html>'
		. '<html><head><title>404 Not Found</title></head>'
		. '<body><h1>Not Found</h1><p>The requested URL was not found on this server.</p>'
		. '</body></html>'
	;

endif;


// Version: 27.1